from __future__ import annotations

from typing import Any

from allotropy.allotrope.converter import unstructure
from allotropy.allotrope.schemas import validate_asm_schema
from allotropy.exceptions import AllotropeSerializationError


def serialize_and_validate_allotrope(model: Any) -> dict[str, Any]:
    try:
        allotrope_dict = unstructure(model)
    except Exception as e:
        msg = f"Failed to serialize allotrope model: {e}"
        raise AllotropeSerializationError(msg) from e

    validate_asm_schema(allotrope_dict)

    return allotrope_dict



"""" Reader file for Thermo Fisher Qubit Flex Parser"""

import numpy as np
import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.pandas import read_excel


class ThermoFisherQubitFlexReader:
    SUPPORTED_EXTENSIONS = "csv,xlsx"

    @classmethod
    def read(cls, named_file_contents: NamedFileContents) -> pd.DataFrame:
        """
        Reads the content of the provided named file and returns it as a pandas DataFrame.

        Parameters:
        named_file_contents (NamedFileContents): The named file contents to read.
            It includes the original file name, file contents, and encoding.

        Returns:
        pd.DataFrame: The content of the file as a pandas DataFrame.

        Raises:
        AllotropeConversionError: If the file format is not supported.
        """
        if named_file_contents.extension == "csv":
            csv_reader = CsvReader.create(named_file_contents)
            # Skip lines until we find the "Run ID" header
            csv_reader.drop_until(r"Test Date")
            # Read the CSV block starting from the header
            df = csv_reader.pop_csv_block_as_df(index_col=False, header=0)
            if df is None:
                msg = "Unable to find 'Run ID' header in CSV file"
                raise AllotropeConversionError(msg)
        else:
            df = read_excel(named_file_contents.contents)
        df = df.replace(np.nan, None)
        return df



from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2025._03.plate_reader import (
    MeasurementGroup,
)
from allotropy.calcdocs.config import (
    CalcDocsConfig,
    CalculatedDataConfig,
    MeasurementConfig,
)
from allotropy.calcdocs.unchained_labs_lunatic_stunner.extractor import LunaticExtractor
from allotropy.calcdocs.unchained_labs_lunatic_stunner.views import (
    DetectionTypeView,
    UuidView,
)
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
)


def create_calculated_data(
    measurement_groups: list[MeasurementGroup],
) -> list[CalculatedDocument]:
    elements = []
    for measurement_group in measurement_groups:
        elements += LunaticExtractor.get_elements(measurement_group.measurements)

    lunatic_view_data = UuidView().apply(elements)

    absorbance_conf = MeasurementConfig(
        name="absorbance",
        value="absorbance",
    )

    dynamic_light_scattering_conf = MeasurementConfig(
        name="dynamic light scattering",
        value="absorbance",
    )

    detection_type_view = DetectionTypeView(sub_view=UuidView()).apply(elements)

    configs = CalcDocsConfig(
        [
            # a260
            CalculatedDataConfig(
                name="Concentration",
                value="a260 concentration (ng/ul)",
                unit="ng/µL",
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            CalculatedDataConfig(
                name="Concentration",
                value="concentration (ng/ul)",
                unit="ng/µL",
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            CalculatedDataConfig(
                name="Background (A260)",
                value="background (a260)",
                unit="mAU",
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            CalculatedDataConfig(
                name="A260/A230",
                value="a260/a230",
                unit=UNITLESS,
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            CalculatedDataConfig(
                name="A260/A280",
                value="a260/a280",
                unit=UNITLESS,
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            # a280
            CalculatedDataConfig(
                name="Concentration",
                value="concentration (mg/ml)",
                unit="mg/mL",
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            CalculatedDataConfig(
                name="Background (A280)",
                value="background (a280)",
                unit="mAU",
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            CalculatedDataConfig(
                name="A260/A280",
                value="a260/a280",
                unit=UNITLESS,
                view_data=lunatic_view_data,
                source_configs=(absorbance_conf,),
            ),
            # Dynamic Light Scattering related calculations
            CalculatedDataConfig(
                name="KC/R (mol/g)",
                value="kc/r (mol/g)",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="kD (ml/g)",
                value="kd (ml/g)",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="B22",
                value="b22 (ml.mol/g^2)",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Optical contrast constant K",
                value="optical contrast constant k (m^2 mol kg^-2)",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Z Average Diameter",
                value="z ave. dia (nm)",
                unit="nm",
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Polydispersity Index",
                value="pdi",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Diameter Standard Deviation",
                value="sd dia (nm)",
                unit="nm",
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Diffusion coefficient",
                value="diffusion coefficient (um^2/s)",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Peak of Interest Mean Diameter",
                value="peak of interest mean dia (nm)",
                unit="nm",
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Peak of Interest Mode Diameter",
                value="peak of interest mode dia (nm)",
                unit="nm",
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Peak of Interest Est. MW",
                value="peak of interest est. mw (kda)",
                unit="kDa",
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Peak of Interest Intensity",
                value="peak of interest intensity (%)",
                unit="%",
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Peak of Interest Mass",
                value="peak of interest mass (%)",
                unit="%",
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Peak of Interest Diffusion coefficient",
                value="peak of interest diffusion coefficient (um^2/s)",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Derived intensity (cps)",
                value="derived intensity (cps)",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Rayleigh ratio R",
                value="rayleigh ratio r (cm^-1)",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="kD goodness of fit",
                value="kd goodness of fit",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
                description_value_key="kd linear fit",
            ),
            CalculatedDataConfig(
                name="B22 goodness of fit",
                value="b22 goodness of fit",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
                description_value_key="b22 linear fit",
            ),
            CalculatedDataConfig(
                name="Viscosity at T (cP)",
                value="viscosity at t (cp)",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Viscosity at 20°C (cP)",
                value="viscosity at 20°c (cp)",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="RI at T",
                value="ri at t",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="RI at 20°C",
                value="ri at 20°c",
                unit=UNITLESS,
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
            CalculatedDataConfig(
                name="Diameter @ C=0",
                value="diameter @ c=0 (nm)",
                unit="nm",
                view_data=detection_type_view,
                source_configs=(dynamic_light_scattering_conf,),
            ),
        ]
    )

    return [
        calc_doc
        for parent_calc_doc in configs.construct()
        for calc_doc in parent_calc_doc.iter_struct()
    ]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import re
from typing import Any, TypedDict

import numpy as np
import pandas as pd

from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2025._03.plate_reader import (
    Data,
    ErrorDocument,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.exceptions import AllotropyParserError
from allotropy.parsers.constants import NEGATIVE_ZERO, NOT_APPLICABLE
from allotropy.parsers.thermo_skanit.constants import DEVICE_TYPE, SAMPLE_ROLE_MAPPINGS
from allotropy.parsers.utils.pandas import df_to_series_data, parse_header_row
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import try_float_or_none

MEASUREMENT_TYPES = {
    MeasurementType.ULTRAVIOLET_ABSORBANCE: "Absorbance",
    MeasurementType.FLUORESCENCE: "Fluorescence",
    MeasurementType.LUMINESCENCE: "Luminescence",
}

SHEET_TABS = [
    "Session information",
    "Instrument information",
    "Layout definitions",
    "General information",
]

GENERAL_INFO_KEYS = [
    "software_name",
    "software_version",
]


class PlateDict(TypedDict):
    wavelength_data: dict[float, pd.DataFrame]
    sample_data: pd.DataFrame


@dataclass(frozen=True)
class PlateData:
    plate_identifier: str
    wavelength_data: dict[float, pd.DataFrame]
    sample_data: pd.DataFrame

    @staticmethod
    def create(
        plate_identifier: str,
        wavelength_data: dict[float, pd.DataFrame],
        sample_data: pd.DataFrame,
    ) -> PlateData:
        return PlateData(
            plate_identifier=plate_identifier,
            wavelength_data=wavelength_data,
            sample_data=sample_data,
        )


@dataclass(frozen=True)
class ThermoSkanItMetadata:
    @staticmethod
    def create_metadata(
        file_path: str,
        instrument_info_df: pd.DataFrame | None,
        general_info_df: pd.DataFrame | None,
    ) -> Metadata:
        instrument_info_data = ThermoSkanItMetadata._get_instrument_data(
            instrument_info_df
        )
        general_info_data = ThermoSkanItMetadata._get_general_info_data(general_info_df)
        path = Path(file_path)

        device_identifier = str(instrument_info_data.pop("device_identifier"))
        model_number = str(instrument_info_data.pop("model_number"))

        equipment_serial_number = None
        if "equipment_serial_number" in instrument_info_data:
            equipment_serial_number = str(
                instrument_info_data.pop("equipment_serial_number")
            )

        software_name = None
        if general_info_data.get("software_name") is not None:
            software_name = str(general_info_data.pop("software_name"))
        else:
            general_info_data.pop("software_name", None)

        software_version = None
        if general_info_data.get("software_version") is not None:
            software_version = str(general_info_data.pop("software_version"))
        else:
            general_info_data.pop("software_version", None)

        # Merge custom_info dicts, ensuring we only unpack dict types
        instrument_custom_info = instrument_info_data.get("custom_info", {})
        device_control_custom_info_raw = instrument_info_data.get(
            "device_control_custom_info"
        )
        general_custom_info = general_info_data.get("custom_info", {})

        custom_info: dict[str, Any] = {}
        if isinstance(instrument_custom_info, dict):
            custom_info.update(instrument_custom_info)
        if isinstance(general_custom_info, dict):
            custom_info.update(general_custom_info)

        # Ensure device_control_custom_info is the correct type
        device_control_custom_info = (
            device_control_custom_info_raw
            if isinstance(device_control_custom_info_raw, dict)
            else None
        )

        return Metadata(
            asm_file_identifier=path.with_suffix(".json").name,
            data_system_instance_id=NOT_APPLICABLE,
            file_name=path.name,
            unc_path=file_path,
            device_identifier=device_identifier,
            model_number=model_number,
            equipment_serial_number=equipment_serial_number,
            software_name=software_name,
            software_version=software_version,
            metadata_custom_info=custom_info,
            device_control_custom_info=device_control_custom_info,
        )

    @staticmethod
    def _get_general_info_data(
        general_info_df: pd.DataFrame | None,
    ) -> dict[str, str | None | dict[str, str]]:
        if general_info_df is None:
            return dict.fromkeys(GENERAL_INFO_KEYS, None)

        general_info_data = df_to_series_data(parse_header_row(general_info_df.T))
        software_info = general_info_data.get(str, "Report generated with SW version")
        if not software_info:
            return dict.fromkeys(GENERAL_INFO_KEYS, NOT_APPLICABLE)
        software_name, software_version_txt = software_info.split(",", 1)
        match = re.search(r"\d+(?:\.\d+)*", software_version_txt)
        software_version = match.group() if match else None
        unread_data = general_info_data.get_unread(skip={"Filter", "Position"})
        # Filter out "nan" keys from custom info and ensure string values
        custom_info = {
            k: str(v) for k, v in unread_data.items() if k != "nan" and v is not None
        }
        return {
            "software_name": software_name,
            "software_version": software_version,
            "custom_info": custom_info,
        }

    @staticmethod
    def _get_instrument_data(
        instrument_info_df: pd.DataFrame | None,
    ) -> dict[str, str | None | dict[str, str] | dict[str, float | str | None]]:
        if instrument_info_df is None:
            return {
                "device_identifier": NOT_APPLICABLE,
                "model_number": NOT_APPLICABLE,
            }

        # Replace empty with "" so we can add label columns together
        instrument_info_df = instrument_info_df.fillna("")

        # The labels for data is spread across the first two columns for some reason, combine them as index.
        # NOTE: This is an assumption that may not be true for future files
        # Combine the first two columns into a Series
        combined_series = instrument_info_df.iloc[:, 0] + instrument_info_df.iloc[:, 1]
        # Convert the Series to an Index
        instrument_info_df.index = pd.Index(combined_series)

        # Read data from the last column, this is where values are found
        instrument_info_data = df_to_series_data(instrument_info_df.T, index=-1)

        lookups = {
            "device_identifier": ("Name", NOT_APPLICABLE),
            "model_number": ("Name", NOT_APPLICABLE),
            "equipment_serial_number": ("Serial number", None),
        }
        # Get the raw data first
        raw_data = {
            key: value
            for key, (label, default) in lookups.items()
            if (value := instrument_info_data.get(str, label, default)) is not None
        }
        device_control_custom_info = instrument_info_data.get_custom_keys(
            {
                "Optical response compensation",
                "Plate adapter number",
                "Plate adapter name",
            }
        )
        unread_data = instrument_info_data.get_unread(
            skip={
                "Wavelength",
                "Incubator",
                "Date and time of definition",
                "Top optics",
                "Bottom optics",
                "Dispenser 1",
                "Dispenser 2",
                "Module's name",
                "Module's serial number",
                "Instrument modules",
                "Gas control",
                "Bandwidth",
                "Position",
                "Filter",
            }
        )
        custom_info = {
            k: str(v) for k, v in unread_data.items() if k != "nan" and v is not None
        }
        return {
            **raw_data,
            "custom_info": custom_info,
            "device_control_custom_info": device_control_custom_info,
        }


@dataclass
class DataWell(Measurement):
    @staticmethod
    def create(
        well_location: str,
        value: float,
        sample_name: str,
        type_: MeasurementType,
        detector_wavelength: float | None,
        well_plate_identifier: str | None,
        error_documents: list[ErrorDocument] | None = None,
        experimental_data_identifier: str | None = None,
        device_control_custom_info: dict[str, Any] | None = None,
    ) -> Measurement:
        measurement_type_str = MEASUREMENT_TYPES[type_]
        error_docs = error_documents or []

        return Measurement(
            type_=type_,
            identifier=random_uuid_str(),
            sample_identifier=sample_name,
            location_identifier=well_location,
            sample_role_type=DataWell.get_sample_role(sample_name),
            detector_wavelength_setting=detector_wavelength,
            device_type=DEVICE_TYPE,
            detection_type=measurement_type_str,
            well_plate_identifier=well_plate_identifier,
            absorbance=value
            if measurement_type_str
            == MEASUREMENT_TYPES[MeasurementType.ULTRAVIOLET_ABSORBANCE]
            else None,
            fluorescence=value
            if measurement_type_str == MEASUREMENT_TYPES[MeasurementType.FLUORESCENCE]
            else None,
            luminescence=value
            if measurement_type_str == MEASUREMENT_TYPES[MeasurementType.LUMINESCENCE]
            else None,
            experimental_data_identifier=experimental_data_identifier,
            error_document=error_docs if error_docs else None,
            device_control_custom_info=device_control_custom_info,
        )

    @staticmethod
    def get_sample_role(sample_name: str) -> SampleRoleType | None:
        stripped_text = re.sub(r"\d+", "", sample_name)
        try:
            return SAMPLE_ROLE_MAPPINGS[stripped_text]
        except KeyError as _:
            return None


@dataclass(frozen=True)
class ThermoSkanItMeasurementGroups:
    @staticmethod
    def create(
        sheet_df: pd.DataFrame,
        type_: MeasurementType,
        session_info_df: pd.DataFrame | None,
        device_control_custom_info: dict[str, Any] | None = None,
    ) -> list[MeasurementGroup]:
        plates = ThermoSkanItMeasurementGroups.identify_data_and_sample_dfs(sheet_df)

        session_name = exec_time = None
        custom_info = {}
        if session_info_df is not None:
            session_info_data = df_to_series_data(parse_header_row(session_info_df.T))
            session_name = session_info_data.get(str, "Session name")
            exec_time = session_info_data.get(str, "Execution time")
            unread_data = session_info_data.get_unread(skip={"Executed with"})
            # Filter out "nan" keys from custom info
            custom_info = {k: v for k, v in unread_data.items() if k != "nan"}

        if not exec_time:
            exec_time = sheet_df.iloc[1].iloc[0]

        if not exec_time:
            msg = "Execution time not found"
            raise AllotropyParserError(msg)

        if not session_name:
            experiment = sheet_df.iloc[0].iloc[0]
            session_name = experiment.replace(".skax", "") if experiment else None

        well_measurements: dict[tuple[str, str], list[Measurement]] = {}
        plate_well_counts: dict[str, int] = {}

        for plate in plates:
            for wavelength, data_df in plate.wavelength_data.items():
                if data_df.empty:
                    continue

                valid_rows = ~pd.isnull(data_df.index) & (data_df.index != "")
                valid_cols = ~pd.isnull(data_df.columns) & (data_df.columns != "")
                plate_well_count = len(data_df.index[valid_rows]) * len(
                    data_df.columns[valid_cols]
                )

                data_df.dropna(how="all", inplace=True)

                stacked = data_df.stack()

                for well_letter, well_column in stacked.index:
                    well_location = well_letter + str(well_column)
                    well_key = (plate.plate_identifier, well_location)

                    if not plate.sample_data.empty:
                        sample_name = plate.sample_data.loc[well_letter, well_column]
                    else:
                        sample_name = f"{plate.plate_identifier}_{well_location}"

                    measurement_value = stacked.loc[well_letter, well_column]
                    error_docs = []

                    if not try_float_or_none(measurement_value):
                        error_value = str(measurement_value)
                        error_docs.append(
                            ErrorDocument(
                                error=error_value,
                                error_feature=MEASUREMENT_TYPES[type_],
                            )
                        )
                        measurement_value = NEGATIVE_ZERO

                    well = DataWell.create(
                        well_location=well_location,
                        value=measurement_value,
                        sample_name=sample_name,
                        detector_wavelength=wavelength if wavelength > 0 else None,
                        type_=type_,
                        well_plate_identifier=plate.plate_identifier,
                        error_documents=error_docs if error_docs else None,
                        experimental_data_identifier=session_name.replace(".skax", "")
                        if session_name
                        else None,
                        device_control_custom_info=device_control_custom_info,
                    )

                    if well_key not in well_measurements:
                        well_measurements[well_key] = []
                    well_measurements[well_key].append(well)

            plate_well_counts[plate.plate_identifier] = plate_well_count

        meas_groups = []
        for (plate_id, _), measurements in well_measurements.items():
            meas_groups.append(
                MeasurementGroup(
                    measurements=measurements,
                    plate_well_count=plate_well_counts[plate_id],
                    measurement_time=exec_time,
                    custom_info=custom_info,
                )
            )

        return meas_groups

    @staticmethod
    def _set_headers(df: pd.DataFrame) -> pd.DataFrame:
        # Set the first column (well letters) as the index
        df.set_index(df.columns[0], inplace=True)
        # Set the first row (well numbers) as the columns
        df = parse_header_row(df)
        valid_columns = pd.notna(df.columns) & (
            df.columns.astype(str).str.lower() != "nan"
        )
        df = df[df.columns[valid_columns]]
        # Cast row numbers to int (float first to handle decimals, e.g. 1.0)
        df.columns = df.columns.astype(float).astype(int)
        return df

    @staticmethod
    def identify_data_and_sample_dfs(
        absorbance_sheet_df: pd.DataFrame,
    ) -> list[PlateData]:
        """Parse multiple plates from sheet data, each with multiple wavelengths."""
        plates_dict: dict[str, PlateDict] = {}
        current_plate_id = None
        current_wavelength = None
        current_data_section: list[pd.Series[Any]] = []
        current_sample_section: list[pd.Series[Any]] = []
        reading_data = False
        reading_samples = False

        for _idx, row in absorbance_sheet_df.iterrows():
            row_str = str(row.iloc[0]) if pd.notna(row.iloc[0]) else ""

            if "Wavelength:" in row_str:
                wavelength_match = re.search(r"Wavelength:\s*(\d{1,3})\s*nm", row_str)
                if wavelength_match:
                    if (
                        current_wavelength is not None
                        and current_data_section
                        and current_plate_id
                    ):
                        data_df = ThermoSkanItMeasurementGroups._set_headers(
                            pd.DataFrame(current_data_section)
                        )
                        if current_plate_id not in plates_dict:
                            plates_dict[current_plate_id] = {
                                "wavelength_data": {},
                                "sample_data": pd.DataFrame(),
                            }
                        wavelength_data = plates_dict[current_plate_id][
                            "wavelength_data"
                        ]
                        wavelength_data[current_wavelength] = data_df

                    current_wavelength = float(wavelength_match.group(1))
                    current_data_section = []
                    reading_data = False
                    reading_samples = False
                    continue

            if "Plate" in row_str:
                plate_match = re.search(r"Plate\s*(\d+)", row_str)
                if plate_match:
                    current_plate_id = plate_match.group(0)
                    if current_plate_id not in plates_dict:
                        plates_dict[current_plate_id] = {
                            "wavelength_data": {},
                            "sample_data": pd.DataFrame(),
                        }
                    continue

            if "Abs" in row.values.astype(str) or "RLU" in row.values.astype(str):
                reading_data = True
                reading_samples = False
                current_data_section = [row]
                continue

            if "Sample" in row.values.astype(str):
                if (
                    current_wavelength is not None
                    and current_data_section
                    and current_plate_id
                ):
                    data_df = ThermoSkanItMeasurementGroups._set_headers(
                        pd.DataFrame(current_data_section)
                    )
                    wavelength_data = plates_dict[current_plate_id]["wavelength_data"]
                    wavelength_data[current_wavelength] = data_df

                reading_samples = True
                reading_data = False
                current_sample_section = [row]
                continue

            if reading_data:
                current_data_section.append(row)
            elif reading_samples:
                current_sample_section.append(row)

                if pd.isna(row.iloc[0]) or str(row.iloc[0]).strip() == "":
                    if current_sample_section and current_plate_id:
                        sample_df = ThermoSkanItMeasurementGroups._set_headers(
                            pd.DataFrame(current_sample_section[:-1])
                        )
                        plates_dict[current_plate_id]["sample_data"] = sample_df
                        current_sample_section = []
                        reading_samples = False

        if current_wavelength is not None and current_data_section and current_plate_id:
            data_df = ThermoSkanItMeasurementGroups._set_headers(
                pd.DataFrame(current_data_section)
            )
            wavelength_data = plates_dict[current_plate_id]["wavelength_data"]
            wavelength_data[current_wavelength] = data_df

        if current_sample_section and current_plate_id:
            sample_df = ThermoSkanItMeasurementGroups._set_headers(
                pd.DataFrame(current_sample_section)
            )
            plates_dict[current_plate_id]["sample_data"] = sample_df

        plates = []
        for plate_id, plate_info in plates_dict.items():
            plates.append(
                PlateData.create(
                    plate_identifier=plate_id,
                    wavelength_data=plate_info["wavelength_data"],
                    sample_data=plate_info["sample_data"],
                )
            )

        return plates


@dataclass(frozen=True)
class DataThermoSkanIt(Data):
    @staticmethod
    def _clean_dataframe(df: pd.DataFrame) -> pd.DataFrame:
        # Prevent pandas from silently downcasting values, to prevent future incompatibility.
        pd.set_option("future.no_silent_downcasting", True)  # noqa: FBT003
        df = df.replace(r"^\s*$", np.nan, regex=True)
        df = df.dropna(axis="index", how="all")
        df = df.dropna(axis="columns", how="all")
        return df

    @staticmethod
    def create(sheet_data: dict[str, pd.DataFrame], file_path: str) -> Data:
        measurement_df = _type = None
        clean_data = {
            key: DataThermoSkanIt._clean_dataframe(sheet_data[key])
            for key in sheet_data.keys()
            if key in SHEET_TABS
        }
        for measurement_type, measurement_type_key in MEASUREMENT_TYPES.items():
            for sheet_name, df in sheet_data.items():
                if measurement_type_key in sheet_name:
                    measurement_df = df
                    _type = measurement_type
                    break

        if measurement_df is None or measurement_df.empty or _type is None:
            msg = "Unsupported or missing measurement type"
            raise AllotropyParserError(msg)
        metadata = ThermoSkanItMetadata.create_metadata(
            instrument_info_df=clean_data.get("Instrument information"),
            general_info_df=clean_data.get("General information"),
            file_path=file_path,
        )

        # Extract device control custom info from metadata
        device_control_custom_info = metadata.device_control_custom_info

        measurement_groups = ThermoSkanItMeasurementGroups.create(
            sheet_df=measurement_df,
            session_info_df=clean_data.get("Session information"),
            type_=_type,
            device_control_custom_info=device_control_custom_info,
        )

        return Data(
            metadata=metadata,
            measurement_groups=measurement_groups,
        )



from allotropy.calcdocs.docs.example.parser_code.measurement import Measurement
from allotropy.calcdocs.extractor import Element, Extractor


class ExampleExtractor(Extractor[Measurement]):
    @classmethod
    def to_element(cls, measurement: Measurement) -> Element:
        return Element(
            uuid=measurement.uuid,
            data={
                "sid": measurement.sid,
                "tid": measurement.tid,
                "m": measurement.m,
                "sum": measurement.sum_,
                "mean": measurement.mean,
            },
        )



from allotropy.allotrope.models.adm.binding_affinity_analyzer.wd._2024._12.binding_affinity_analyzer import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.binding_affinity_analyzer.benchling._2024._12.binding_affinity_analyzer import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.cytiva_biacore_t200_evaluation import constants
from allotropy.parsers.cytiva_biacore_t200_evaluation.cytiva_biacore_t200_evaluation_data_creator import (
    create_data as _create_data,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class CytivaBiacoreT200EvaluationParser(VendorParser[Data, Model]):
    DISPLAY_NAME = constants.DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "bme"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        metadata, groups = _create_data(named_file_contents)
        return Data(metadata=metadata, measurement_groups=groups)



from io import StringIO

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_gen5.agilent_gen5_structure import (
    Gen5DataContext,
    get_concentrations,
    get_identifiers,
    get_kinetic_measurements,
    get_temperature,
    HeaderData,
    KineticData,
    ReadData,
)
from allotropy.parsers.agilent_gen5.constants import (
    MULTIPLATE_FILE_ERROR,
    NO_MEASUREMENTS_ERROR,
    NO_PLATE_DATA_ERROR,
)
from allotropy.parsers.lines_reader import SectionLinesReader
from allotropy.parsers.utils.pandas import df_to_series_data, read_csv, SeriesData
from allotropy.parsers.utils.values import assert_not_none


class AgilentGen5Reader:
    header_data: SeriesData
    sections: dict[str, list[str]]
    SUPPORTED_EXTENSIONS = "txt"

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = SectionLinesReader.create(named_file_contents)

        plate_readers = list(reader.iter_sections("^Software Version"))
        if not plate_readers:
            raise AllotropeConversionError(NO_PLATE_DATA_ERROR)

        if len(plate_readers) > 1:
            raise AllotropeConversionError(MULTIPLATE_FILE_ERROR)

        plate_reader = plate_readers[0]

        # Header data is all data until the "Procedure Details" section, read in as a cvs
        assert_not_none(
            plate_reader.drop_until("^Software Version"), "Software Version"
        )
        lines = [line for line in plate_reader.pop_until("Procedure Details") if line]
        df = read_csv(
            StringIO("\n".join(lines)),
            header=None,
            index_col=0,
            keep_default_na=False,
            sep="\t",
        ).T
        self.header_data = df_to_series_data(df)

        self.sections = {}

        # Special handling for the "Procedure Details" section, which has an empty line after the title.
        assert_not_none(
            plate_reader.drop_until("^Procedure Details"), "Procedure Details"
        )
        plate_reader.pop()
        plate_reader.drop_empty()
        # add the plate type to the header data to extract the well count if needed
        self.header_data.series["Plate Type"] = plate_reader.get()
        self.sections["Procedure Details"] = list(plate_reader.pop_until_empty())
        plate_reader.drop_empty()

        while plate_reader.current_line_exists():
            lines = list(plate_reader.pop_until_empty())
            self.sections[lines[0].split("\t")[0].strip(":")] = lines
            plate_reader.drop_empty()

    def get_required_section(self, section_name: str) -> list[str]:
        """Get a required section, raises error if not found."""
        section = self.sections.get(section_name)
        if section is None:
            msg = f"Required section '{section_name}' not found."
            raise AllotropeConversionError(msg)
        return section

    @property
    def procedure_details(self) -> list[str]:
        return self.get_required_section("Procedure Details")

    @property
    def layout_section(self) -> list[str] | None:
        return self.sections.get("Layout")

    @property
    def wavelength_section(self) -> list[str] | None:
        return self.sections.get("Wavelength")

    @property
    def time_section(self) -> list[str] | None:
        return self.sections.get("Time")

    @property
    def actual_temperature_section(self) -> list[str] | None:
        return self.sections.get("Actual Temperature")

    def _validate_result_sections(self, result_sections: list[list[str]]) -> None:
        """Validates whether all the result sections dimensions are consistent."""
        first_section = result_sections[0]

        for section in result_sections[1:]:
            if not first_section[0] == section[0] and len(first_section) == len(
                section
            ):
                msg = "All result tables should have the same dimensions."
                raise AllotropeConversionError(msg)

    def get_results_section(self) -> list[str] | None:
        """Returns a valid Results Matrix from the reader sections if found.

        Checks for Results in the reader sections, if not found, creates the results matrix with all
        sections that are correctly formatted as a results table (excluding the Layout section). If
        no tables with results are found, returns None
        """
        if "Results" in self.sections:
            return self.sections["Results"]

        def is_results(section: list[str]) -> bool:
            return (
                len(section) > 2
                and section[1].startswith("\t1")
                and section[2].startswith("A\t")
            )

        result_sections = []
        for name, section in self.sections.items():
            if name == "Layout":
                continue
            if is_results(section):
                result_sections.append(section[1:])

        if result_sections:
            self._validate_result_sections(result_sections)
            return [
                "Results",
                result_sections[0][0],
                *[
                    section[i + 1]
                    for i in range(len(result_sections[0]) - 1)
                    for section in result_sections
                ],
            ]

        return None

    def extract_data_context(self, file_path: str) -> Gen5DataContext:
        results_section = self.get_results_section()
        if not results_section:
            self.header_data.get_unread()
            raise AllotropeConversionError(NO_MEASUREMENTS_ERROR)

        kinetic_result = get_kinetic_measurements(self.time_section)
        kinetic_measurements, kinetic_elapsed_time, kinetic_errors = kinetic_result or (
            {},
            [],
            {},
        )

        return Gen5DataContext(
            header_data=HeaderData.create(self.header_data, file_path),
            read_data=ReadData.create(self.procedure_details),
            kinetic_data=KineticData.create(self.procedure_details),
            results_section=results_section,
            sample_identifiers=get_identifiers(self.layout_section),
            concentration_values=get_concentrations(self.layout_section),
            actual_temperature=get_temperature(self.actual_temperature_section),
            kinetic_measurements=kinetic_measurements,
            kinetic_elapsed_time=kinetic_elapsed_time,
            kinetic_errors=kinetic_errors,
            wavelength_section=self.wavelength_section,
        )



from collections import defaultdict

from allotropy.calcdocs.extractor import Element
from allotropy.calcdocs.view import View


class AssayIdentifierView(View):
    def __init__(self, sub_view: View | None = None):
        super().__init__(name="assay_identifier", sub_view=sub_view)

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        items = defaultdict(list)
        for element in elements:
            assay_identifier = element.get_str("assay_identifier")
            items[assay_identifier].append(element)
        return dict(items)


class SampleIdentifierView(View):
    def __init__(self, sub_view: View | None = None):
        super().__init__(name="sample_identifier", sub_view=sub_view)

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        items = defaultdict(list)
        for element in elements:
            sample_identifier = element.get_str("sample_identifier")
            items[sample_identifier].append(element)
        return dict(items)


class WellPlateIdentifierView(View):
    def __init__(self, sub_view: View | None = None):
        super().__init__(name="well_plate_identifier", sub_view=sub_view)

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        items = defaultdict(list)
        for element in elements:
            well_plate_identifier = element.get_str("well_plate_identifier")
            items[well_plate_identifier].append(element)
        return dict(items)


class LocationIdentifierView(View):
    def __init__(self, sub_view: View | None = None):
        super().__init__(name="location_identifier", sub_view=sub_view)

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        items = defaultdict(list)
        for element in elements:
            location_identifier = element.get_str("location_identifier")
            items[location_identifier].append(element)
        return dict(items)



from typing import Any

from allotropy.calcdocs.extractor import Element, Extractor
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import (
    WellItem,
)


def get_value(extra_data: dict[str, Any] | None, key: str) -> Any | None:
    return extra_data.get(key) if extra_data else None


class AppbioQuantstudioExtractor(Extractor[WellItem]):
    @classmethod
    def to_element(cls, well_item: WellItem) -> Element:
        return Element(
            uuid=well_item.uuid,
            data={
                "uuid": well_item.uuid,
                "identifier": well_item.identifier,
                "target_dna_description": well_item.target_dna_description,
                "sample_identifier": well_item.sample_identifier,
                "location_identifier": well_item.location_identifier,
                "reporter_dye_setting": well_item.reporter_dye_setting,
                "position": well_item.position,
                "well_location_identifier": well_item.well_location_identifier,
                "quencher_dye_setting": well_item.quencher_dye_setting,
                "sample_role_type": (
                    well_item.sample_role_type.value
                    if well_item.sample_role_type
                    else None
                ),
                "group_identifier": well_item.group_identifier,
                "well_identifier": get_value(well_item.extra_data, "well identifier"),
                "sample_color": get_value(well_item.extra_data, "sample color"),
                "biogroup_color": get_value(well_item.extra_data, "biogroup color"),
                "target_color": get_value(well_item.extra_data, "target color"),
                "cycle_threshold_value_setting": well_item.result.cycle_threshold_value_setting,
                "cycle_threshold_result": well_item.result.cycle_threshold_result,
                "automatic_cycle_threshold_enabled_setting": well_item.result.automatic_cycle_threshold_enabled_setting,
                "automatic_baseline": well_item.result.automatic_baseline,
                "baseline_start": well_item.result.baseline_start,
                "baseline_end": well_item.result.baseline_end,
                "normalized_reporter_result": well_item.result.normalized_reporter_result,
                "baseline_corrected_reporter_result": well_item.result.baseline_corrected_reporter_result,
                "genotyping_determination_result": well_item.result.genotyping_determination_result,
                "genotyping_determination_method_setting": well_item.result.genotyping_determination_method_setting,
                "quantity": well_item.result.quantity,
                "quantity_mean": well_item.result.quantity_mean,
                "quantity_sd": well_item.result.quantity_sd,
                "ct_mean": well_item.result.ct_mean,
                "ct_sd": well_item.result.ct_sd,
                "delta_ct_mean": well_item.result.delta_ct_mean,
                "delta_ct_se": well_item.result.delta_ct_se,
                "delta_delta_ct": well_item.result.delta_delta_ct,
                "rq": well_item.result.rq,
                "rq_min": well_item.result.rq_min,
                "rq_max": well_item.result.rq_max,
                "rn_mean": well_item.result.rn_mean,
                "rn_sd": well_item.result.rn_sd,
                "y_intercept": well_item.result.y_intercept,
                "r_squared": well_item.result.r_squared,
                "slope": well_item.result.slope,
                "efficiency": well_item.result.efficiency,
                "comments": well_item.result.comments,
                "amp_score": well_item.result.amp_score,
                "cq_conf": well_item.result.cq_conf,
                "omit": get_value(well_item.result.extra_data, "omit"),
                "highsd": get_value(well_item.result.extra_data, "highsd"),
                "noamp": get_value(well_item.result.extra_data, "noamp"),
                "expfail": get_value(well_item.result.extra_data, "expfail"),
                "tholdfail": get_value(well_item.result.extra_data, "tholdfail"),
                "prfdrop": get_value(well_item.result.extra_data, "prfdrop"),
            },
        )



from decimal import Decimal
from pathlib import Path

from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    Error,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.parsers.constants import DEFAULT_EPOCH_TIMESTAMP, NOT_APPLICABLE
from allotropy.parsers.revvity_matrix import constants
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_metadata(file_path: str, headers: SeriesData) -> Metadata:
    path = Path(file_path)
    return Metadata(
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_id=NOT_APPLICABLE,
        device_identifier=NOT_APPLICABLE,
        file_name=path.name,
        unc_path=file_path,
        device_type=constants.DEVICE_TYPE,
        software_name="Revvity Matrix",
        software_version=headers.get(str, "Version"),
        equipment_serial_number=headers.get(str, "Instrument SN"),
    )


def create_measurement_group(data: SeriesData, headers: SeriesData) -> MeasurementGroup:
    # This function will be called for every row in the dataset, use it to create
    # a corresponding measurement group.

    viable_cell_density = data[float, ["Live Concentration (E6)", "Live Cells/mL"]]
    # Live Cells/mL needs to be converted to 10^6 cells/mL
    if not data.has_key("Live Concentration (E6)"):
        viable_cell_density = float(Decimal(viable_cell_density) / Decimal("1000000"))

    total_cell_density = data.get(float, ["Total Concentration (E6)", "Total Cells/mL"])
    # Total Cells/mL needs to be converted to 10^6 cells/mL
    if total_cell_density is not None and not data.has_key("Total Concentration (E6)"):
        total_cell_density = float(Decimal(total_cell_density) / Decimal("1000000"))

    dead_cell_density = data.get(float, ["Dead Concentration (E6)", "Dead Cells/mL"])
    # Dead Cells/mL needs to be converted to 10^6 cells/mL
    if dead_cell_density is not None and not data.has_key("Dead Concentration (E6)"):
        dead_cell_density = float(Decimal(dead_cell_density) / Decimal("1000000"))

    errors = data.get(str, ["Errors:", "Errors"], validate=SeriesData.NOT_NAN)

    return MeasurementGroup(
        analyst=headers.get(str, "Operator"),
        custom_info={
            "assay identifier": headers.get(str, "Assay Name"),
        },
        measurements=[
            Measurement(
                measurement_identifier=random_uuid_str(),
                # NOTE: instrument file does not provide a timestamp, but it is required by ASM, so pass
                # EPOCH to signal no timestamp.
                timestamp=headers.get(
                    str, "Results Timestamp", default=DEFAULT_EPOCH_TIMESTAMP
                ),
                sample_identifier=data[str, "Well Name"],
                viability=data[float, ["Viability", "Viability (%)"]],
                total_cell_count=data.get(float, "Total Count"),
                total_cell_density=total_cell_density,
                average_total_cell_diameter=data.get(
                    float, ["Total Mean Size", "Total Diameter"]
                ),
                viable_cell_count=data.get(float, "Live Count"),
                viable_cell_density=viable_cell_density,
                average_live_cell_diameter=data.get(
                    float, ["Live Mean Size", "Live Diameter"]
                ),
                dead_cell_count=data.get(float, "Dead Count"),
                dead_cell_density=dead_cell_density,
                average_dead_cell_diameter=data.get(
                    float, ["Dead Mean Size", "Dead Diameter"]
                ),
                errors=[
                    Error(error=error, feature="Cell Counting")
                    for error in (errors.split(",") if errors else [])
                ],
                sample_custom_info={
                    "Row": data.get(str, "Row"),
                    "Column": data.get(str, "Column"),
                    "Well Plate Identifier": headers.get(str, "Plate Name"),
                },
                processed_data_identifier=random_uuid_str(),
                cell_aggregation_percentage=data.get(float, "Aggregates (%)"),
                aggregate_count=data.get(float, "Aggregate Count"),
                aggregate_size=data.get(float, "Aggregate Size"),
                cell_density_dilution_factor=headers.get(float, "Dilution"),
                custom_info={
                    "Scanresult Timestamp": headers.get(str, "Scanresult Timestamp"),
                    **(data.get_unread()),
                    **(headers.get_unread()),
                },
            )
        ],
    )



from typing import IO

import chardet

from allotropy.constants import CHARDET_ENCODING, DEFAULT_ENCODING
from allotropy.exceptions import AllotropeConversionError, AllotropeParsingError


def _get_contents(contents: bytes | IO[bytes] | IO[str]) -> str | bytes:
    if isinstance(contents, bytes):
        return contents
    actual_contents = contents.read()
    contents.seek(0)
    return actual_contents


def determine_encoding(
    contents: bytes | IO[bytes] | IO[str], encoding: str | None
) -> list[str | None]:
    if not encoding:
        return [DEFAULT_ENCODING]
    if encoding != CHARDET_ENCODING:
        return [encoding]

    actual_contents = _get_contents(contents)
    if isinstance(actual_contents, str):
        return [None]
    if not actual_contents:
        msg = "Unable to detect encoding for empty bytes string, file may be empty."
        raise AllotropeConversionError(msg)

    detect_result = chardet.detect(actual_contents)
    if not detect_result["encoding"]:
        msg = (
            f"Unable to detect text encoding for file with content: {actual_contents!r}"
        )
        raise AllotropeParsingError(msg)
    # chardet can report the wrong encoding when there are strange characters in the contents (e.g. emojis)
    # To address this, we take the following approach - if the confidence of the detection is < 70%, report
    # DEFAULT_ENCODING first, and the detected encoding second. If we return multiple encodings, the caller
    # should try all.
    if detect_result["confidence"] < 0.7:
        return [DEFAULT_ENCODING, detect_result["encoding"]]
    return [detect_result["encoding"]]


def decode(contents: IO[bytes] | IO[str], encoding: str | None) -> str:
    actual_contents = _get_contents(contents)
    if isinstance(actual_contents, str):
        return actual_contents
    possible_encodings = determine_encoding(actual_contents, encoding)

    for encoding in possible_encodings:
        # NOTE: this should not be possible, we only return None if contents is str, which should have returned already.
        if encoding is None:
            msg = f"Could not determine encoding of contents: {actual_contents!r}"
            raise AssertionError(msg)
        try:
            return actual_contents.decode(encoding)
        except UnicodeDecodeError as e:
            if encoding != possible_encodings[-1]:
                continue
            msg = f"Unable to decode bytes with encoding '{encoding}' with error: {e}, bytes: {actual_contents!r}"
            raise AllotropeParsingError(msg) from e
        except LookupError as e:
            msg = f"Invalid encoding: '{encoding}'."
            raise AllotropeConversionError(msg) from e
    msg = f"Unable to decode contents with possible encodings: {possible_encodings}. Contents: {actual_contents!r}"
    raise AllotropeConversionError(msg)



from __future__ import annotations

from pathlib import Path

from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.parsers.beckman_vi_cell_xr.constants import (
    DEFAULT_ANALYST,
    DETECTION_TYPE,
    DEVICE_TYPE,
    MODEL_NUMBER,
    SOFTWARE_NAME,
)
from allotropy.parsers.beckman_vi_cell_xr.vi_cell_xr_reader import ViCellData
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_measurement_group(data: SeriesData) -> MeasurementGroup:
    total_cell_count = data.get(float, "Total cells")
    total_cell_count = (
        total_cell_count if total_cell_count is None else round(total_cell_count)
    )
    viable_cell_count = data.get(float, "Viable cells")
    viable_cell_count = (
        viable_cell_count if viable_cell_count is None else round(viable_cell_count)
    )

    custom_info_measurement_group = data.get_custom_keys(
        {
            "experimental data identifier": ["File Name", "File name"],
            "log date": ["LogDate", "Log date"],
            "file date": ["FileDate", "File date"],
        }
    )

    # Mark unused keys to be ignored
    data.mark_read({"Settings:", "Results:", "Settings", "Results"})

    return MeasurementGroup(
        analyst=DEFAULT_ANALYST,
        measurements=[
            Measurement(
                measurement_identifier=random_uuid_str(),
                timestamp=data[str, "Sample date"],
                sample_identifier=data[str, "Sample ID"],
                cell_type_processing_method=data.get(str, "Cell type"),
                cell_density_dilution_factor=data.get(float, "Dilution factor"),
                viability=data[float, "Viability (%)"],
                viable_cell_density=data[float, "Viable cells/ml (x10^6)"],
                total_cell_count=total_cell_count,
                total_cell_density=data.get(float, "Total cells/ml (x10^6)"),
                average_total_cell_diameter=data.get(float, "Avg. diam. (microns)"),
                viable_cell_count=viable_cell_count,
                dead_cell_count=data.get(float, "Nonviable cells"),
                average_total_cell_circularity=data.get(float, "Avg. circ."),
                maximum_cell_diameter_setting=data.get(float, "Max diam. (microns)")
                or data.get(float, "Maximum diameter (microns)"),
                minimum_cell_diameter_setting=data.get(float, "Min diam. (microns)")
                or data.get(float, "Minimum diameter (microns)"),
                sample_custom_info=data.get_custom_keys(
                    {
                        "Internal Dilution",
                        "Dilution",
                    }
                ),
                device_control_custom_info=data.get_custom_keys(
                    {
                        "Aspirate cycles",
                        "Probe volume (ml x 10^-6)",
                        "Trypan blue mixing cycles",
                    }
                ),
                data_processing_custom_info=data.get_custom_keys(
                    {
                        "Decluster degree",
                        "Viable cell spot area (%)",
                        "Viable spot area",
                        "Viable cell spot brightness (%)",
                        "V. cell spot brightness (%)",
                        "Cell sharpness",
                        "Sharpness",
                        "Cell brightness (%)",
                        "Brightness (%)",
                        "Brightness",
                        "Analysis version",
                        "Number of bins",
                        "Minimum circularity",
                        "Sample depth (microns)",
                    }
                ),
                image_processing_custom_info=data.get_custom_keys(
                    {
                        "Images",
                        "Saved images",
                        "ImageBaseName",
                        "Frames",
                        "Microns/pixel ratio",
                        "ImageDirectory",
                        "Image size",
                        "Field of view (microns)",
                    }
                ),
                processed_data_custom_info=data.get_custom_keys(
                    {
                        "Avg. cells / image",
                        "Average cells / image",
                        "Background intensity sum",
                        "Total diameter sum",
                        "Total circularity sum",
                        "Avg. background intensity",
                        "Avg. bg intensity",
                    }
                ),
                custom_info=data.get_unread(),
            )
        ],
        custom_info=custom_info_measurement_group,
    )


def create_metadata(reader_data: ViCellData, file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        device_type=DEVICE_TYPE,
        detection_type=DETECTION_TYPE,
        model_number=MODEL_NUMBER,
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_id=NOT_APPLICABLE,
        device_identifier=NOT_APPLICABLE,
        equipment_serial_number=reader_data.serial_number,
        software_name=SOFTWARE_NAME,
        software_version=reader_data.version.value,
        file_name=path.name,
        unc_path=file_path,
    )



from functools import partial
from pathlib import Path

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.liquid_handler.benchling._2024._11.liquid_handler import (
    Error,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.beckman_echo_plate_reformat import constants
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_metadata(header_footer_data: SeriesData, file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        file_name=path.name,
        asm_file_identifier=path.with_suffix(".json").name,
        unc_path=str(path),
        data_system_instance_identifier=NOT_APPLICABLE,
        device_type=constants.DEVICE_TYPE,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        software_version=header_footer_data[str, "Application Version"],
        model_number=header_footer_data[str, "Instrument Model"],
        equipment_serial_number=header_footer_data[str, "Instrument Serial Number"],
        software_name=header_footer_data[str, "Application Name"],
        device_system_custom_info={
            "Instrument Software Version": header_footer_data.get(
                str, "Instrument Software Version"
            )
        },
        custom_info=header_footer_data.get_unread(),
    )


def _create_measurement(row_data: SeriesData, run_date_time: str | None) -> Measurement:
    def convert_echo_nl_to_ul(value: float | None) -> float | None:
        return (
            (value * constants.PLATE_REFORMAT_REPORT_VOLUME_CONVERSION_TO_UL)
            if value is not None
            else None
        )

    # If the Date Time Point only contains time (no date) combine with Run Date/Time to create measurement time.
    # We detect if there is a date in the timestamp by checking for a space, and if there is no space, add the date
    # component of
    date_time_point = row_data[str, "Date Time Point"]
    if " " not in date_time_point.strip():
        if not run_date_time or " " not in run_date_time:
            msg = "Cannot parse timestamp for measurement, 'Date Time Point' does not contain a date (time only) and there is no valid 'Run Date/Time' in the header to infer date from."
            raise AllotropeConversionError(msg)
        run_date_time_date = run_date_time.split(" ")
        date_time_point = f"{run_date_time_date[0]} {date_time_point}"

    return Measurement(
        identifier=random_uuid_str(),
        measurement_time=date_time_point,
        sample_identifier=row_data.get(str, "Sample ID", NOT_APPLICABLE),
        source_plate=row_data[str, "Source Plate Barcode"],
        source_well=row_data[str, "Source Well"],
        source_location=row_data[str, "Source Plate Name"],
        destination_plate=row_data[str, "Destination Plate Barcode"],
        destination_well=row_data[str, "Destination Well"],
        destination_location=row_data[str, "Destination Plate Name"],
        aspiration_volume=convert_echo_nl_to_ul(row_data.get(float, "Actual Volume")),
        transfer_volume=convert_echo_nl_to_ul(row_data.get(float, "Actual Volume")),
        injection_volume_setting=convert_echo_nl_to_ul(
            row_data.get(float, "Transfer Volume")
        ),
        device_control_custom_info={
            "sample name": row_data.get(str, "Sample Name"),
            "survey fluid volume": row_data.get(
                float, "Survey Fluid Volume"
            ),  # This is already in uL, so don't convert to nL
            "current fluid volume": row_data.get(
                float, "Current Fluid Volume"
            ),  # This is already in uL, so don't convert to nL
            "intended transfer volume": convert_echo_nl_to_ul(
                row_data.get(float, "Transfer Volume")
            ),
            "source labware name": row_data.get(str, "Source Plate Type"),
            "destination labware name": row_data.get(str, "Destination Plate Type"),
            "fluid composition": row_data.get(str, "Fluid Composition"),
            "fluid units": row_data.get(str, "Fluid Units"),
            "fluid type": row_data.get(str, "Fluid Type"),
            "transfer status": row_data.get(str, "Transfer Status", NOT_APPLICABLE),
        },
        errors=[
            Error(
                error=row_data[str, "Transfer Status"],
                feature=row_data[str, "Transfer Status"].split(": ")[0],
            )
        ]
        if row_data.get(str, "Transfer Status")
        else [],
        custom_info=row_data.get_unread(),
    )


def create_measurement_groups(
    data: pd.DataFrame, header: SeriesData
) -> list[MeasurementGroup]:
    run_date_time = header.get(str, "Run Date/Time")
    create_measurement = partial(_create_measurement, run_date_time=run_date_time)
    return [
        MeasurementGroup(
            analyst=header[str, "User Name"],
            analytical_method_identifier=header.get(str, "Protocol Name"),
            experimental_data_identifier=header.get(str, "Run ID"),
            measurements=map_rows(data, create_measurement),
            custom_info={
                "Run Date/Time": run_date_time,
            },
        )
    ]



""" Decodes a zipped .rslt folder file into intermediate json"""
import os
from pathlib import Path
import re
import tempfile
from typing import Any, IO
import zipfile

import rainbow.agilent.chemstation as rb  # type: ignore
import xmltodict


def merge_peak_with_signal_name(
    peak_data: list[dict[str, Any]],
    signal_data: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    """
    Matches the signal IDs from the injections with those in the peaks and combines them accordingly.
    :param peak_data: peak data
    :param signal_data: signal id data
    :return: combined peak data
    """
    signal_id_to_name = {
        signal_id: signal_value
        for item in signal_data
        for signal_id, signal_value in item.items()
        if signal_id != "Name"
    }

    for peak in peak_data:
        signal_id = peak["Signal_ID"]
        if signal_id in signal_id_to_name:
            peak["Signal_ID"] = signal_id_to_name[signal_id]

    return peak_data


def decode_data_cubes(
    datacube_file: IO[bytes], datacube_file_name: str
) -> dict[str, Any]:
    """
    decodes the detection data to datacubes
    :param datacube_path: detection file path
    :return: decoded datacube
    """
    with tempfile.NamedTemporaryFile(
        "wb", suffix=Path(datacube_file_name).suffix
    ) as fp:
        fp.write(datacube_file.read())
        fp.seek(0)
        raw_data = rb.parse_ch(fp.name)
    if ".CH" in os.path.basename(datacube_file_name):
        datacubes = {
            "Time": [time * 60 for time in raw_data.xlabels],
            "Intensity": [float(sublist[0]) for sublist in raw_data.data],
            "Metadata": raw_data.metadata,
            "Chromatogram filename": os.path.basename(datacube_file_name),
        }
    else:
        datacubes = {
            "Time": [time * 60 for time in raw_data.xlabels],
            "Pressure": [float(sublist[0]) / 10 for sublist in raw_data.data],
            "Metadata": raw_data.metadata,
            "Chromatogram filename": os.path.basename(datacube_file_name),
        }
    return datacubes


def extract_rx_file(rx_file: IO[bytes]) -> list[dict[str, Any]]:
    """
    Extracts the peak and its associated metadata from the rx file for each injection.
    :param temporary_input_path: rx file to be extracted
    :param zip_file_path: temporary unzip path
    :param injection_signal: each injection signal data
    :return: peak details along with its metadata
    """
    peak_details: list[dict[str, Any]] = []
    with zipfile.ZipFile(rx_file, "r") as zip_ref:
        processed_file_path = next(
            iter(_get_matching_filenames(zip_ref, ".*InjectionACAML"))
        )
        with zip_ref.open(processed_file_path) as processed_file:
            peaks = xmltodict.parse(processed_file.read().decode("utf-8-sig"))
            for peak in peaks["ACAML"]["Doc"]["Content"]["Injections"]["Result"][
                "SignalResult"
            ]:
                peak_data = {}
                if "Peak" in peak:
                    peak_data["Peak"] = peak["Peak"]
                    peak_data["Signal_ID"] = peak["Signal_ID"]["@id"]
                    peak_details.append(peak_data)

            peak_metadata_dict = {}
            if isinstance(
                peaks["ACAML"]["Doc"]["Content"]["Injections"]["Result"][
                    "InjectionCompound"
                ],
                list,
            ):
                for item in peaks["ACAML"]["Doc"]["Content"]["Injections"]["Result"][
                    "InjectionCompound"
                ]:
                    peak_id = item.get("@id")
                    if peak_id:
                        identification = item.get("Identification", None)
                        identification_id = (
                            identification.get("Qualified", {})
                            .get("Peaks", {})
                            .get("Peak_ID", {})
                            .get("@id")
                        )
                        if identification_id:
                            peak_metadata_dict[identification_id] = item
            else:
                peak_id = (
                    peaks["ACAML"]["Doc"]["Content"]["Injections"]["Result"][
                        "InjectionCompound"
                    ]["Identification"]
                    .get("Qualified", {})
                    .get("Peaks", {})
                    .get("Peak_ID", {})
                    .get("@id")
                )
                if peak_id:
                    peak_metadata_dict[peak_id] = peaks["ACAML"]["Doc"]["Content"][
                        "Injections"
                    ]["Result"]["InjectionCompound"]

            if isinstance(peak_details, list):
                for peak_dict in peak_details:
                    if isinstance(peak_dict.get("Peak"), dict):
                        peak_id = peak_dict["Peak"]["@id"]
                        if peak_id in peak_metadata_dict:
                            peak_dict["Peak"]["Peak Metadata"] = peak_metadata_dict[
                                peak_id
                            ]
                    elif isinstance(peak_dict.get("Peak"), list):
                        for peak in peak_dict["Peak"]:
                            peak_id = peak["@id"]
                            if peak_id in peak_metadata_dict:
                                peak["Peak Metadata"] = peak_metadata_dict[peak_id]

    return peak_details


def extract_dx_file(
    dx_file: IO[bytes],
    injection_metadata_data: dict[str, Any],
    dx_file_name: str,
) -> list[dict[str, Any]]:
    """
    temporarily unzips the dx file and decodes the data and structures it
    :param temporary_input_path: temporary path to unzip file
    :param zip_file_path: compressed dx zip path
    :param injection_metadata_data: metadata of injection
    :return: chromatogram data, pump pressure data and its metadata for each injection
    """

    sample_data: dict[str, Any] = {}
    chromatogram_data: list[dict[str, Any]] = []
    with zipfile.ZipFile(dx_file, "r") as zip_ref:
        acmd_filepath = next(iter(_get_matching_filenames(zip_ref, r".*\.acmd")))
        with zip_ref.open(acmd_filepath) as injection_file_data:
            acmd_data = injection_file_data.read().decode("utf-8-sig")
            injection_data = xmltodict.parse(acmd_data)
            injection_data["ACMD"]["InjectionInfo"].pop("Signals")
            for sample_setup in injection_metadata_data["SampleSetup"]:
                if (
                    sample_setup["IdentParam"]["Name"]
                    in injection_data["ACMD"]["InjectionInfo"]["SampleName"]
                ):
                    sample_data["SampleSetup"] = sample_setup
            for sample_measurement in injection_metadata_data["SampleMeasurement"]:
                if (
                    sample_measurement["IdentParam"]["Name"]
                    in injection_data["ACMD"]["InjectionInfo"]["SampleName"]
                ):
                    sample_data["SampleMeasurement"] = sample_measurement

            sample_data["sequence_data"] = injection_metadata_data["sequence_data"]
            sample_data.update(injection_data["ACMD"]["InjectionInfo"])

            sample_data.update(injection_metadata_data["sequence_data"])
        chrom_files = _get_matching_filenames(
            zip_ref, r".*\.CH"
        ) + _get_matching_filenames(
            zip_ref, rf".*{injection_metadata_data['pump_pressure_filename']}.*"
        )
        for chrom_file_name in chrom_files:
            with zip_ref.open(chrom_file_name) as chrom_file:
                chromatogram_data.append(decode_data_cubes(chrom_file, chrom_file_name))

    for each_chromatogram in chromatogram_data:
        each_chromatogram["Sample Data"] = sample_data
        each_chromatogram["file_name"] = dx_file_name
    return chromatogram_data


def extract_sqx_file(sqx_file: IO[bytes]) -> dict[str, Any]:
    """
    unzips the sqx file and retrieves analysis method path
    :param temporary_input_path: temporary unzip file path
    :param zip_file_path: sqx file path
    :return: structured analysis method data
    """
    with zipfile.ZipFile(sqx_file, "r") as zip_ref:
        sequence_sample_data = {}
        sample_file_path = next(
            iter(_get_matching_filenames(zip_ref, ".*SampleListPart$"))
        )
        with zip_ref.open(sample_file_path) as sample_file:
            sequence_sample_data["AnalysisMethod"] = xmltodict.parse(
                sample_file.read()
            )["SequenceTable"]["anyType"][0]["AnalysisMethod"]
            return sequence_sample_data


def decode_acaml_data(
    acaml_content: dict[str, Any]
) -> tuple[dict[str, Any], dict[str, Any], dict[str, Any]]:
    """
    structures the acaml file data
    :param acaml_content: decoded acaml data
    :return: structured acaml content
    """
    metadata_data: dict[str, Any] = {}
    instrument_data: dict[str, Any] = {}
    pump_pressure_file: dict[str, Any] = {}
    signal_details: dict[str, Any] = {"Signal details": []}

    content_data = acaml_content["ACAML"]["Doc"]["Content"]
    instrument_fields = [
        "Name",
        "Technique",
        "@id",
        "Module",
        "AcquisitionApplication",
        "CreatedByUser",
        "CreationDate",
    ]
    for key in instrument_fields:
        if key in content_data["Resources"]["Instrument"]:
            instrument_data[key] = content_data["Resources"]["Instrument"][key]
        elif key in content_data["Injections"]["MeasData"][0]:
            instrument_data[key] = content_data["Injections"]["MeasData"][0][key]
        elif key in acaml_content["ACAML"]["Doc"]["DocInfo"]:
            instrument_data[key] = acaml_content["ACAML"]["Doc"]["DocInfo"][key]
    for measurements in content_data["Injections"]["MeasData"]:
        signal_data = {}
        for signals in measurements["Signal"]:
            signal_data["Name"] = signals["BinaryData"]["DataItem"]["Name"].split(".")[
                0
            ]
            signal_data[signals["@id"]] = signals["Name"]
            if "Pressure" in signals["Description"]:
                pump_pressure_file[
                    signals["BinaryData"]["DataItem"]["Name"].split(".")[0]
                ] = signals["TraceID"]
        signal_details["Signal details"].append(signal_data)
    metadata_data["Instrument"] = instrument_data
    metadata_data["SeparationMedium"] = content_data["Resources"]["SeparationMedium"]
    metadata_data["SampleSetup"] = content_data["Samples"]["Setup"]
    metadata_data["SampleMeasurement"] = content_data["Samples"]["MeasData"]

    return metadata_data, pump_pressure_file, signal_details


def _get_matching_filenames(zip_ref: zipfile.ZipFile, pattern: str) -> list[str]:
    return [
        name
        for name in zip_ref.namelist()
        if not Path(name).name.startswith(".") and re.match(pattern, name)
    ]


def decode_data(input_bytes: IO[bytes]) -> dict[str, Any]:
    """
    decoded the files in input folder path and returns a structured data
    :param input_path: input folder path
    :return: structured intermediate json
    """
    intermediate_json: dict[str, Any] = {}
    injection_data: dict[str, Any] = {}
    total_injection_chromatogram_details: list[dict[str, Any]] = []
    total_peak_details: list[dict[str, Any]] = []

    with zipfile.ZipFile(input_bytes) as zip_ref:
        acaml_path = next(iter(_get_matching_filenames(zip_ref, ".*acaml")))
        with zip_ref.open(acaml_path) as acaml_file_data:
            acaml_file_content = acaml_file_data.read().decode("utf-8-sig")
            decoded_acaml_content = xmltodict.parse(acaml_file_content)
        (
            intermediate_metadata,
            pump_pressure_file,
            injection_signal_details,
        ) = decode_acaml_data(decoded_acaml_content)
        injection_data["total_pressure_files"] = pump_pressure_file
        injection_data["SampleSetup"] = intermediate_metadata["SampleSetup"]
        injection_data["SampleMeasurement"] = intermediate_metadata["SampleMeasurement"]
        injection_data["signal_details"] = injection_signal_details
        intermediate_metadata.pop("SampleSetup")
        intermediate_metadata.pop("SampleMeasurement")
        intermediate_json["Metadata"] = intermediate_metadata

        sequence_file_name = next(iter(_get_matching_filenames(zip_ref, r".*\.sqx")))
        with zip_ref.open(sequence_file_name, "r") as sequence_file:
            sequence_data = extract_sqx_file(sequence_file)
            injection_data["sequence_data"] = sequence_data

        dx_files = _get_matching_filenames(zip_ref, r".*\.dx")
        for dx_file_name in dx_files:
            name_without_ext = Path(dx_file_name).with_suffix("").name
            if name_without_ext in injection_data["total_pressure_files"]:
                injection_data["pump_pressure_filename"] = injection_data[
                    "total_pressure_files"
                ][name_without_ext]
            with zip_ref.open(dx_file_name) as dx_file:
                total_injection_chromatogram_details.extend(
                    extract_dx_file(dx_file, injection_data, dx_file_name)
                )

        rx_files = _get_matching_filenames(zip_ref, r".*\.rx")
        for rx_file_name in rx_files:
            with zip_ref.open(rx_file_name) as rx_file:
                processed_peak_data: dict[str, Any] = {}
                peak_details = extract_rx_file(rx_file)
                peak_details = merge_peak_with_signal_name(
                    peak_details, injection_data["signal_details"]["Signal details"]
                )
                processed_peak_data["file_name"] = rx_file_name
                processed_peak_data["peak_details"] = peak_details
                total_peak_details.append(processed_peak_data)

        for injection_details in total_injection_chromatogram_details:
            for peaks_data in total_peak_details:
                if (
                    Path(injection_details["file_name"]).stem
                    == Path(peaks_data["file_name"]).stem
                ):
                    peak_details_data = peaks_data["peak_details"]
                    for peak_detail in peak_details_data:
                        if isinstance(peak_detail, dict):
                            signal_name = injection_details["Metadata"]["signal"].split(
                                ","
                            )[0]
                            peak_signal = peak_detail["Signal_ID"]
                            if peak_signal == signal_name:
                                if isinstance(peak_detail["Peak"], list):
                                    injection_details.update(peak_detail)
                                if isinstance(peak_detail["Peak"], dict):
                                    peak_detail["Peak"] = [peak_detail["Peak"]]
                                    injection_details.update(peak_detail)

        total_injection_chromatogram_details.sort(
            key=lambda x: x["Chromatogram filename"]
        )
        intermediate_json["Result Data"] = total_injection_chromatogram_details
        intermediate_json["Sample Count"] = {
            "count": len(total_injection_chromatogram_details)
        }
        return intermediate_json



from allotropy.allotrope.models.adm.binding_affinity_analyzer.wd._2024._12.binding_affinity_analyzer import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.binding_affinity_analyzer.benchling._2024._12.binding_affinity_analyzer import (
    Data as MapperData,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.cytiva_biacore_insight.constants import DISPLAY_NAME
from allotropy.parsers.cytiva_biacore_insight.cytiva_biacore_insight_data_creator import (
    create_calculated_data,
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.cytiva_biacore_insight.cytiva_biacore_insight_reader import (
    CytivaBiacoreInsightReader,
)
from allotropy.parsers.cytiva_biacore_insight.cytiva_biacore_insight_structure import (
    Data,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class CytivaBiacoreInsightParser(VendorParser[MapperData, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = CytivaBiacoreInsightReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> MapperData:
        reader = CytivaBiacoreInsightReader.create(named_file_contents)
        data = Data.create(reader)
        return MapperData(
            create_metadata(data.metadata, named_file_contents.original_file_path),
            create_measurement_groups(data),
            create_calculated_data(data),
        )



import pandas as pd

from allotropy.allotrope.models.adm.liquid_handler.benchling._2024._11.liquid_handler import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.liquid_handler.benchling._2024._11.liquid_handler import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_echo_plate_reformat.beckman_echo_plate_reformat_reader import (
    BeckmanEchoPlateReformatReader,
)
from allotropy.parsers.beckman_echo_plate_reformat.beckman_echo_plate_reformat_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.beckman_echo_plate_reformat.constants import DISPLAY_NAME
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class BeckmanEchoPlateReformatParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = BeckmanEchoPlateReformatReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = BeckmanEchoPlateReformatReader(named_file_contents)
        measurement_groups = create_measurement_groups(
            pd.concat(reader.sections.values(), ignore_index=True), reader.header
        )
        return Data(
            create_metadata(reader.header, named_file_contents.original_file_path),
            measurement_groups,
        )



from pathlib import Path
from typing import Any

from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.models.shared.definitions.units import (
    Microliter,
    MilliAbsorbanceUnit,
    MilliAbsorbanceUnitTimesSecond,
    Milliliter,
    SecondTime,
)
from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    DeviceControlDoc,
    DeviceDocument,
    Measurement,
    MeasurementGroup,
    Metadata,
    Peak,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.benchling_chromeleon import constants
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import try_float, try_float_or_none


def _create_device_documents(
    device_information: dict[str, Any]
) -> list[DeviceDocument] | None:
    return [
        DeviceDocument(
            device_type="pump",
            model_number=device_information.get("pump model number") or NOT_APPLICABLE,
        ),
        DeviceDocument(
            device_type="uv",
            model_number=device_information.get("uv model number") or NOT_APPLICABLE,
        ),
        DeviceDocument(
            device_type="sampler",
            model_number=device_information.get("sampler model number")
            or NOT_APPLICABLE,
        ),
    ]


def create_metadata(
    first_injection: dict[str, Any],
    sequence: dict[str, Any],
    file_path: str,
    device_information: dict[str, Any],
) -> Metadata:
    return Metadata(
        asset_management_identifier=first_injection.get(
            "precondition system instrument name", NOT_APPLICABLE
        ),
        software_name=constants.SOFTWARE_NAME,
        file_name=Path(file_path).name,
        unc_path=file_path,
        description=first_injection.get("description"),
        device_documents=_create_device_documents(device_information),
        lc_agg_custom_info={
            "Sequence Creation Time": sequence.get("sequence creation time"),
            "Sequence Directory": sequence.get("sequence directory"),
            "Sequence Name": sequence.get("sequence name"),
            "Sequence Update operator": sequence.get("sequence update operator"),
            "Sequence Update Time": sequence.get("sequence update time"),
            "Number of Injections": sequence.get("number of injections"),
        },
    )


def _validate_injection_volume_unit(injection_volume: float, unit: str) -> float:
    # if unit is Milliliter, convert to Microliter
    if unit == Milliliter.unit:
        return injection_volume * 1000
    if unit != Microliter.unit:
        msg = f"Invalid injection volume unit: {unit}"
        raise AllotropeConversionError(msg)
    return injection_volume


def _get_chromatogram(signal: dict[str, Any]) -> DataCube | None:
    chrom: dict[str, list[float]] | None = signal.get("chromatogram")
    if not chrom:
        return None
    if len(chrom) != 2:
        msg = "Expected chrom to have two lists"
        raise AllotropeConversionError(msg)

    dimensions = chrom["x"]
    measures = chrom["y"]

    # ASM expected chromatogram dimensions (x axis) to be in seconds, but Chromeleon reports it in minutes,
    # so convert here.
    dimensions = [t * 60 for t in dimensions]

    return DataCube(
        label=signal["signal name"],
        structure_dimensions=[
            DataCubeComponent(
                type_=FieldComponentDatatype.double,
                concept="retention time",
                unit=SecondTime.unit,
            )
        ],
        structure_measures=[
            DataCubeComponent(
                type_=FieldComponentDatatype.double,
                concept="absorbance",
                unit=MilliAbsorbanceUnit.unit,
            )
        ],
        dimensions=[dimensions],
        measures=[measures],
    )


def _convert_to_seconds(value: float | None) -> float | None:
    # Assume value is in minutes
    return value * 60 if value is not None else None


def _create_peak(peak: dict[str, Any], signal: dict[str, Any]) -> Peak:
    # Area and height are reported in μV, but are reported in ASM as mAU
    # For Chromeleon software, 1V == 1AU, so we just need to convert μ to m
    if (area := try_float_or_none(peak.get("area"))) is not None:
        area /= 1000
    if (height := try_float_or_none(peak.get("height"))) is not None:
        height /= 1000
    # Times are reported in minutes by Chromeleon - convert to seconds
    retention_time = _convert_to_seconds(try_float_or_none(peak.get("retention time")))
    width_at_half_height = _convert_to_seconds(
        try_float_or_none(peak.get("peak width at half height"))
    )
    peak_width_at_5_percent_of_height = _convert_to_seconds(
        try_float_or_none(peak.get("peak width at 5 % of height"))
    )
    peak_width_at_10_percent_of_height = _convert_to_seconds(
        try_float_or_none(peak.get("peak width at 10 % of height"))
    )
    peak_width_at_baseline = _convert_to_seconds(
        try_float_or_none(peak.get("peak width at baseline"))
    )
    baseline_value_at_start_of_peak = _convert_to_seconds(
        try_float_or_none(peak.get("start value baseline"))
    )
    baseline_value_at_end_of_peak = _convert_to_seconds(
        try_float_or_none(peak.get("stop value baseline"))
    )
    peak_right_width_at_10_percent_height = _convert_to_seconds(
        try_float_or_none(peak.get("peak right width at 10 % of height"))
    )
    peak_left_width_at_10_percent_height = _convert_to_seconds(
        try_float_or_none(peak.get("peak left width at 10 % of height"))
    )
    peak_group = _convert_to_seconds(try_float_or_none(peak.get("group area")))

    return Peak(
        identifier=random_uuid_str(),
        index=peak.get("identifier"),
        start=try_float(peak.get("start time"), "start time") * 60,
        start_unit=SecondTime.unit,
        end=try_float(peak.get("end time"), "end time") * 60,
        end_unit=SecondTime.unit,
        retention_time=retention_time,
        area=area,
        area_unit=MilliAbsorbanceUnitTimesSecond.unit,
        relative_area=try_float_or_none(peak.get("relative peak area")),
        width=try_float_or_none(peak.get("Width")),
        height=height,
        height_unit="mAU",
        relative_height=try_float_or_none(peak.get("relative peak height")),
        written_name=peak.get("name"),
        relative_retention_time=try_float_or_none(peak.get("relative retention time")),
        capacity_factor=try_float_or_none(peak.get("capacity factor")),
        chromatographic_resolution=try_float_or_none(
            peak.get("chromatographic peak resolution")
        ),
        number_of_theoretical_plates_by_peak_width_at_half_height=try_float_or_none(
            peak.get("number of theoretical plates by peak width at half height")
        ),
        width_at_half_height=width_at_half_height,
        peak_width_at_5_percent_of_height=peak_width_at_5_percent_of_height,
        peak_width_at_10_percent_of_height=peak_width_at_10_percent_of_height,
        peak_width_at_baseline=peak_width_at_baseline,
        asymmetry_factor_measured_at_5_percent_height=try_float_or_none(
            peak.get("asymmetry factor measured at 5 % height")
        ),
        peak_analyte_amount=try_float_or_none(peak.get("amount")),
        relative_corrected_peak_area=try_float_or_none(peak.get("rel ce area total")),
        peak_group=peak_group,
        baseline_value_at_start_of_peak=baseline_value_at_start_of_peak,
        baseline_value_at_end_of_peak=baseline_value_at_end_of_peak,
        custom_info={
            "number of peaks": signal.get("number of peaks"),
            "peak right width at 10% height": peak_right_width_at_10_percent_height,
            "peak left width at 10% height": peak_left_width_at_10_percent_height,
            "chromatographic peak resolution (usp)": peak.get(
                "chromatographic peak resolution (USP)"
            ),
            "asymmetry aia": peak.get("asymmetry aia"),
        },
    )


def _create_measurements(injection: dict[str, Any]) -> list[Measurement] | None:
    injection_volume_setting = injection.get("injection volume setting")
    injection_volume_unit = injection.get("injection volume unit")
    if injection_volume_setting and injection_volume_unit:
        injection_volume_setting = _validate_injection_volume_unit(
            injection_volume_setting, injection_volume_unit
        )

    signals = injection.get("signals", [])
    if len(signals) == 0:
        return None
    return [
        Measurement(
            measurement_identifier=random_uuid_str(),
            description=injection.get("description"),
            sample_identifier=injection["sample identifier"],
            location_identifier=injection.get("location identifier"),
            well_location_identifier=injection.get("custom variables", {}).get("Well"),
            observation=injection.get("custom variables", {}).get("Observation"),
            sample_custom_info={
                "sample precipitation": injection.get("custom variables", {}).get(
                    "Sample_Precipitation"
                ),
                "rack type": injection.get("custom variables", {}).get("Rack_Type"),
                "stability": injection.get("custom variables", {}).get("Stability"),
                "req id": injection.get("custom variables", {}).get("REQ_ID"),
                "additional comment": injection.get("custom variables", {}).get(
                    "Additional_Comment"
                ),
            },
            device_control_docs=[
                DeviceControlDoc(
                    device_type=constants.DEVICE_TYPE,
                    detection_type=signal.get("detection type"),
                    detector_offset_setting=val
                    if (val := signal.get("detector offset setting")) != "unknown"
                    else None,
                    detector_wavelength_setting=signal.get("wavelength setting"),
                    detector_sampling_rate_setting=val
                    if (val := signal.get("detector sampling rate setting"))
                    != "unknown"
                    else None,
                    detector_bandwidth_setting=signal.get("bandwidth setting"),
                    electronic_absorbance_reference_wavelength_setting=signal.get(
                        "reference wavelength setting"
                    ),
                    electronic_absorbance_reference_bandwidth_setting=signal.get(
                        "reference bandwidth setting"
                    ),
                )
            ],
            injection_identifier=injection["injection identifier"],
            injection_time=injection["injection time"],
            injection_volume_setting=injection_volume_setting,
            injection_custom_info={
                "injection": injection.get("injection number"),
                "injection name": injection.get("injection name"),
                "injection position": injection.get("injection position"),
                "injection status": injection.get("injection status"),
                "injection type": injection.get("injection type"),
                "last update user name": injection.get("last update user name"),
                "creation user name": injection.get("creation user name"),
                "injection program": injection.get("injection program"),
                "injection method": injection.get("injection method"),
            },
            chromatography_serial_num=NOT_APPLICABLE,
            chromatogram_data_cube=_get_chromatogram(signal) if signal else None,
            peaks=[_create_peak(peak, signal) for peak in signal.get("peaks", [])],
        )
        for signal in injection.get("signals", [])
    ]


def create_measurement_groups(
    injections: list[dict[str, Any]]
) -> list[MeasurementGroup]:
    return [
        MeasurementGroup(measurements=measurements)
        for sample_injections in injections
        if (measurements := _create_measurements(sample_injections)) is not None
    ]



import json
from typing import Any, cast

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.json import JsonData
from allotropy.parsers.utils.values import assert_not_none


class BenchlingEmpowerReader:
    SUPPORTED_EXTENSIONS = "json"

    metadata_fields: JsonData
    injections: list[JsonData]

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        contents_dict: dict[str, Any] = json.load(named_file_contents.contents)
        contents = JsonData(contents_dict)

        values_data: dict[str, Any] = assert_not_none(
            contents.data.get("values"), "values"
        )
        metadata: dict[str, Any] = contents.data.get("metadata", {})

        contents.mark_read({"values", "metadata"})

        values = JsonData(values_data)

        fields: dict[str, Any] = assert_not_none(
            values.data.get("fields"), "values/fields"
        )

        self.metadata_fields = JsonData(
            {
                **fields,
                **metadata,
                "instrument_methods": values.data.get("instrument_methods", []),
                "processing_methods": values.data.get("processing_methods", []),
            }
        )

        values.mark_read(
            {
                "fields",
                "instrument_methods",
                "processing_methods",
                "injections",
                "channels",
                "results",
            }
        )

        # Each injection corresponds to a measurement document
        injections_list: list[dict[str, Any]] = assert_not_none(
            values.data.get("injections"), "values/injections"
        )

        id_to_injection_data = {}
        for idx, injection_dict in enumerate(injections_list):
            injection = JsonData(injection_dict)
            injection.mark_read("fields")
            fields_dict = injection.data.get("fields")
            if not fields_dict:
                msg = f"Missing 'fields' for injection at index {idx}"
                raise AssertionError(msg)

            injection_id = fields_dict.get("InjectionId")
            if injection_id is None:
                msg = f"Missing 'InjectionId' for injection at index {idx}"
                raise AssertionError(msg)

            id_to_injection_data[injection_id] = fields_dict

        # Channels contain fields and chromatograms
        channels: list[dict[str, Any]] = values.data.get("channels", [])
        for channel_dict in channels:
            channel = JsonData(channel_dict)
            channel_fields_dict: dict[str, Any] = assert_not_none(
                channel.data.get("fields"), "channels/fields"
            )

            inj_id = channel_fields_dict.get("InjectionId")
            if inj_id is None:
                channel_id = channel_fields_dict.get("ChannelId", "Unknown")
                msg = f"Expected InjectionId in 'fields' for channel: {channel_id}"
                raise AssertionError(msg)

            # Update injection fields with channel fields
            for key, value in channel_fields_dict.items():
                if (
                    key in id_to_injection_data[inj_id]
                    and id_to_injection_data[inj_id][key] != value
                ):
                    msg = f"Mismatch between injection field and channel field for key: {key}"
                    raise AssertionError(msg)
                id_to_injection_data[inj_id][key] = value

            # Add chromatogram data if present
            chrom: list[list[float]] | None = channel.data.get("chrom")
            if chrom is not None:
                id_to_injection_data[inj_id]["chrom"] = chrom

            channel.mark_read({"fields", "chrom"})

        # Results contain peaks and calibration curves
        results: list[dict[str, Any]] = values.data.get("results", [])
        for result_dict in results:
            result = JsonData(result_dict)
            result_fields_dict: dict[str, Any] = assert_not_none(
                result.data.get("fields"), "results/fields"
            )

            inj_id = result_fields_dict.get("InjectionId")
            if inj_id is None:
                result_id = result_fields_dict.get("ResultId", "Unknown")
                msg = f"Expected InjectionId in 'fields' for result: {result_id}"
                raise AssertionError(msg)

            if "results" not in id_to_injection_data[inj_id]:
                id_to_injection_data[inj_id]["results"] = []
            id_to_injection_data[inj_id]["results"].append(result_fields_dict)

            # Add peaks if present
            peaks: list[dict[str, Any]] | None = result.data.get("peaks")
            if peaks is not None:
                id_to_injection_data[inj_id]["peaks"] = [
                    cast(dict[str, Any], peak.get("fields", {})) for peak in peaks
                ]

            # Add calibration curves if present
            curves: list[Any] | None = result.data.get("curves")
            if curves is not None:
                id_to_injection_data[inj_id]["curves"] = curves

            result.mark_read({"fields", "peaks", "curves"})

        self.injections = [JsonData(data) for data in id_to_injection_data.values()]



from pathlib import Path

from allotropy.allotrope.schema_mappers.adm.multi_analyte_profiling.benchling._2024._09.multi_analyte_profiling import (
    Calibration,
    Metadata,
)
from allotropy.parsers.luminex_intelliflex.constants import DEFAULT_SOFTWARE_NAME
from allotropy.parsers.luminex_xponent.constants import DEFAULT_DEVICE_TYPE
from allotropy.parsers.luminex_xponent.luminex_xponent_structure import Header


def create_metadata(
    header: Header, calibrations: list[Calibration], file_path: str
) -> Metadata:
    path = Path(file_path)
    return Metadata(
        file_name=path.name,
        asm_file_identifier=path.with_suffix(".json").name,
        unc_path=file_path,
        equipment_serial_number=header.equipment_serial_number,
        model_number=header.model_number,
        calibrations=calibrations,
        data_system_instance_identifier=header.data_system_instance_identifier,
        software_name=DEFAULT_SOFTWARE_NAME,
        software_version=header.software_version,
        device_type=DEFAULT_DEVICE_TYPE,
    )



import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import parse_header_row, read_excel, SeriesData


class BeckmanPharmspecReader:
    SUPPORTED_EXTENSIONS = "xlsx"
    data: pd.DataFrame
    header: SeriesData

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        df = read_excel(named_file_contents.contents, header=None, engine="calamine")

        """
        Find the data in the raw dataframe. We identify the boundary of the data
        by finding the index first row which contains the word 'Particle' and ending right before
        the index of the first row containing 'Approver'.
        """
        # Check for rows containing "Particle" in column 1
        particle_rows = df[df[1].str.contains("Particle", na=False)].index.values
        if len(particle_rows) == 0:
            msg = "Unable to find required 'Particle' marker in column 1 of the data file."
            raise AllotropeConversionError(msg)
        start = particle_rows[0]

        # Check for rows containing "Approver_" in column 0
        approver_rows = df[df[0].str.contains("Approver_", na=False)].index.values
        if len(approver_rows) == 0:
            msg = "Unable to find required 'Approver_' marker in column 0 of the data file."
            raise AllotropeConversionError(msg)
        end = approver_rows[0] - 1

        # The header data is everything up to the start of the data.
        # It is stored in two columns spread over the first 6 columns.
        raw_header = df.loc[: start - 1].T
        header_data = pd.concat([raw_header.loc[2], raw_header.loc[5]])
        header_columns = pd.concat([raw_header.loc[0], raw_header.loc[3]])

        # Store the software version string before filtering (it's usually at index 0)
        software_version_string = (
            str(header_data.iloc[0]) if len(header_data) > 0 else "Unknown"
        )

        # Filter out nan values from header_columns to avoid nan keys in the Series
        valid_mask = ~pd.isna(header_columns)
        header_data = header_data[valid_mask]
        header_columns = header_columns[valid_mask]

        header_data.index = pd.Index(header_columns)

        # Create a SeriesData with the software version string preserved
        series_data = SeriesData(header_data)
        series_data._software_version_string = software_version_string  # type: ignore
        self.header = series_data

        data = df.loc[start:end]
        data = parse_header_row(data)
        data["Run No."] = data["Run No."].ffill()
        self.data = data.dropna(subset="Particle Size(µm)")



DISPLAY_NAME = "Beckman Echo Plate Reformat"
DEVICE_TYPE = "liquid handler"
PRODUCT_MANUFACTURER = "Beckman Coulter"
SOFTWARE_NAME = "Labcyte Echo Plate Reformat"

TRANSFER_RESULTS_LABEL = "DETAILS"
TRANSFER_ERRORS_LABEL = "EXCEPTIONS"

PLATE_REFORMAT_REPORT_VOLUME_UNITS = "nL"
PLATE_REFORMAT_REPORT_VOLUME_CONVERSION_TO_UL = 0.001



from enum import Enum

from allotropy.allotrope.models.shared.definitions.definitions import (
    TStatisticDatumRole,
)

# FlowJo parser constants
DEVICE_IDENTIFIER = "cytometer"
DEVICE_TYPE = "Flow Cytometer"
SOFTWARE_NAME = "FlowJo"
DISPLAY_NAME = "FlowJo"

# Equipment serial number keywords
EQUIPMENT_SERIAL_KEYWORDS = {"CTNUM", "$CYTSN"}

# Keywords extracted at measurement document level (custom_info)
MEASUREMENT_DOCUMENT_KEYWORDS = {
    "$INST",
    "FJ_FCS_VERSION",
    "$BEGINANALYSIS",
    "$ENDANALYSIS",
    "$BEGINSTEXT",
    "$ENDSTEXT",
    "$BEGINDATA",
    "$ENDDATA",
    "$FIL",
    "$SYS",
    "$TOT",
    "$PAR",
    "$MODE",
    "$BYTEORD",
    "$DATATYPE",
    "$NEXTDATA",
    "CREATOR",
    "$SRC",
    "EXPERIMENT NAME",
    "GUID",
    "$DATE",
    "$BTIM",
    "$ETIM",
    "SETTINGS",
    "WINDOW EXTENSION",
    "EXPORT USER NAME",
    "EXPORT TIME",
    "modDate",
    "name",
    "clientTimestamp",
    "homepage",
}

# Keywords extracted at sample document level (custom_info)
SAMPLE_DOCUMENT_KEYWORDS = {"TUBE NAME"}

# Keywords extracted at processed data document level (custom_info)
PROCESSED_DATA_KEYWORDS = {
    "linFromKW",
    "logFromKW",
    "linMax",
    "logMax",
    "useFCS3",
    "useGain",
    "linearRescale",
    "THRESHOLD",
    "APPLY COMPENSATION",
    "$TIMESTEP",
    "FSC ASF",
    "AUTOBS",
    "SPILL",
    "logMin",
    "linMin",
    "extraNegs",
    "logRescale",
}

# Combined set of all structured keywords (for exclusion logic)
ALL_STRUCTURED_KEYWORDS = (
    MEASUREMENT_DOCUMENT_KEYWORDS
    | SAMPLE_DOCUMENT_KEYWORDS
    | PROCESSED_DATA_KEYWORDS
    | EQUIPMENT_SERIAL_KEYWORDS
    | {
        "$OP",
        "WELL ID",
        "PLATE ID",
        "$CYT",
        "CYTNUM",
        "curGroup",
    }  # Other keywords used elsewhere
)

# Map of FlowJo field names to statistic datum roles and units
FLOWJO_STATISTIC_MAP = {
    # Fluorescence statistics
    "Median": {"role": TStatisticDatumRole.median_role.value, "unit": "RFU"},
    "CV": {
        "role": TStatisticDatumRole.coefficient_of_variation_role.value,
        "unit": "%",
    },
    "Robust CV": {
        "role": TStatisticDatumRole.robust_coefficient_of_variation_role.value,
        "unit": "%",
    },
    "Mean": {"role": TStatisticDatumRole.arithmetic_mean_role.value, "unit": "RFU"},
    "Geometric Mean": {
        "role": TStatisticDatumRole.geometric_mean_role.value,
        "unit": "RFU",
    },
    "Percentile": {"role": TStatisticDatumRole.percentile_role.value, "unit": "RFU"},
    "SD": {
        "role": TStatisticDatumRole.standard_deviation_role.value,
        "unit": "(unitless)",
    },
    "MADExact": {
        "role": TStatisticDatumRole.median_absolute_deviation_percentile_role.value,
        "unit": "(unitless)",
    },
    "Robust SD": {
        "role": TStatisticDatumRole.robust_standard_deviation_role.value,
        "unit": "(unitless)",
    },
    "Median Abs Dev": {
        "role": TStatisticDatumRole.median_absolute_deviation_role.value,
        "unit": "(unitless)",
    },
    # Count statistics
    "fj.stat.freqofparent": {
        "role": TStatisticDatumRole.frequency_of_parent_role.value,
        "unit": "%",
    },
    "fj.stat.freqofgrandparent": {
        "role": TStatisticDatumRole.frequency_of_grandparent_role.value,
        "unit": "%",
    },
    "fj.stat.freqoftotal": {
        "role": TStatisticDatumRole.frequency_of_total_role.value,
        "unit": "%",
    },
}

# Identify statistics that belong to the "Count" feature
COUNT_FIELDS = [
    "fj.stat.freqofparent",
    "fj.stat.freqofgrandparent",
    "fj.stat.freqoftotal",
    "fj.stat.freqof",
]

FLUORESCENCE_FIELDS = [
    "Median",
    "CV",
    "Robust CV",
    "Mean",
    "Geometric Mean",
    "Percentile",
    "SD",
    "MADExact",
    "Robust SD",
    "Median Abs Dev",
]


class RegionType(Enum):
    RECTANGLE = "Rectangle"
    POLYGON = "Polygon"
    ELLIPSOID = "Ellipsoid"
    CURLY_QUAD = "CurlyQuad"


class VertexRole(Enum):
    FOCI = "foci"
    EDGE = "edge"



from dataclasses import dataclass

from allotropy.allotrope.models.shared.definitions.definitions import (
    TStatisticDatumRole,
)

DEFAULT_SOFTWARE_NAME = "xPONENT"
DEFAULT_CONTAINER_TYPE = "well plate"
DEFAULT_DEVICE_TYPE = "multi analyte profiling analyzer"

LUMINEX_EMPTY_PATTERN = r"^[,\"\s]*$"
CALIBRATION_BLOCK_HEADER = "Most Recent Calibration and (Control|Verification) Results"
TABLE_HEADER_PATTERN = '^"?DataType:"?,"?([^,"]*)"?'
MINIMUM_CALIBRATION_LINE_COLS = 2
EXPECTED_CALIBRATION_RESULT_LEN = 2
EXPECTED_HEADER_COLUMNS = 7
REQUIRED_SECTIONS = ["Count", "Units", "Dilution Factor"]
POSSIBLE_WELL_COUNTS_LUMINEX = [96, 384]


@dataclass(frozen=True)
class StatisticSectionConfig:
    role: TStatisticDatumRole
    unit: str


STATISTIC_SECTIONS_CONF = {
    "Median": StatisticSectionConfig(
        role=TStatisticDatumRole.median_role,
        unit="RFU",
    ),
    "Mean": StatisticSectionConfig(
        role=TStatisticDatumRole.arithmetic_mean_role,
        unit="RFU",
    ),
    "%CV": StatisticSectionConfig(
        role=TStatisticDatumRole.coefficient_of_variation_role,
        unit="(unitless)",
    ),
    "Standard Deviation": StatisticSectionConfig(
        role=TStatisticDatumRole.standard_deviation_role,
        unit="(unitless)",
    ),
    "Peak": StatisticSectionConfig(
        role=TStatisticDatumRole.mode_value_role,
        unit="RFU",
    ),
    "Trimmed Peak": StatisticSectionConfig(
        role=TStatisticDatumRole.trimmed_mode_value_role,
        unit="RFU",
    ),
    "Trimmed Count": StatisticSectionConfig(
        role=TStatisticDatumRole.trimmed_count_role,
        unit="RFU",
    ),
    "Trimmed Standard Deviation": StatisticSectionConfig(
        role=TStatisticDatumRole.trimmed_standard_deviation_role,
        unit="(unitless)",
    ),
    "Trimmed Mean": StatisticSectionConfig(
        role=TStatisticDatumRole.trimmed_arithmetic_mean_role,
        unit="RFU",
    ),
}

CALCULATED_DATA_SECTIONS: dict[str, str] = {
    "Net MFI": "RFU",
    "Avg Net MFI": "RFU",
    "Test Result": "(unitless)",
    "% Recovery": "(unitless)",
    "%CV of Replicates": "(unitless)",
}

# Known metric names in the columns of the single dataset results table
SINGLE_DATASET_RESULTS_METRIC_WORDS: set[str] = {
    "MEAN",
    "MEDIAN",
    "COUNT",
    "PEAK",
    "STDEV",
    "STANDARD",
    "DEVIATION",
    "%CV",
    "TRIMMED",
    "NET",
    "MFI",
    "AVERAGE",
    "AVG",
    "RMS",
    "MODE",
    "TRMEAN",
    "TRIMSTDEV",
    "NORMALIZED",
    "DILUTION",
    "FACTOR",
    "SETTING",
    "UNITS",
}

# Allowed section names for Luminex Xponent reports. Any other name should raise an error upstream.
ALLOWED_SECTION_NAMES: set[str] = {
    "Median",
    "Test Result",
    "Range",
    "Net MFI",
    "Count",
    "Mean",
    "Avg Net MFI",
    "Peak",
    "Trimmed Peak",
    "Trimmed Count",
    "Trimmed Std Dev",
    "Std Dev",
    "% CV",
    "Expected Result",
    "Units",
    "Per Bead Count",
    "Dilution Factor",
    "Analysis Types",
    "Audit Logs",
    "Warnings/Errors",
}



import csv
from io import StringIO

import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.utils.pandas import read_csv, SeriesData
from allotropy.parsers.utils.values import assert_not_none


class BmgMarsReader:
    SUPPORTED_EXTENSIONS = "csv"
    header: SeriesData
    data: pd.DataFrame
    header_content: str

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        try:
            self._parse_file(named_file_contents)
        except pd.errors.ParserError:
            # Fix malformed CSV files with uneven columns and retry
            named_file_contents = self._normalize_csv_columns(named_file_contents)
            self._parse_file(named_file_contents)

    def _parse_file(self, named_file_contents: NamedFileContents) -> None:
        """Parse the BMG MARS file and populate header and data attributes."""
        reader = CsvReader(read_to_lines(named_file_contents))
        lines = list(reader.pop_until_inclusive("^,?Raw Data"))
        # Store the header contents so we can parse some values that don't have key/value
        # pairs such as wavelengths and read type.
        self.header_content = "\n".join(lines)

        # Transform header data into a single series
        raw_data = read_csv(StringIO(self.header_content), header=None)
        df = pd.melt(raw_data, value_vars=raw_data.columns.to_list()).dropna(
            axis="index"
        )
        new = df["value"].str.split(": ", expand=True, n=1)
        # Handle the case where no ": " delimiter is found, resulting in a DataFrame with only one column
        if new.shape[1] < 2:
            msg = "Unable to parse header data: no key-value pairs found with expected format."
            raise AllotropeConversionError(msg)
        self.header = SeriesData(pd.Series(new[1].values, index=new[0].str.upper()))

        # Read in the rest of the file as a dataframe
        reader.drop_empty(r"^[,\"\s]*$")
        self.data = assert_not_none(
            reader.pop_csv_block_as_df(header=0, index_col=0),
            msg="Unable to parse dataset from file.",
        )

    def _normalize_csv_columns(
        self, named_file_contents: NamedFileContents
    ) -> NamedFileContents:
        """Fix malformed CSV files with uneven number of columns per row.

        BMG MARS software <= v4.0 can produce CSV files where some rows have
        more or fewer columns than others. This method normalizes all rows
        to have the same number of columns by padding with empty strings,
        allowing the csv reader to parse the file correctly.

        Args:
            named_file_contents: The named file contents to fix

        Returns:
            Fixed NamedFileContents with consistent column counts
        """
        lines = read_to_lines(named_file_contents)

        if not lines:
            return named_file_contents

        # Parse all rows and find the maximum number of columns
        csv_rows = []
        max_columns = 0

        for line in lines:
            if line.strip():
                reader = csv.reader([line])
                try:
                    row = next(reader)
                    csv_rows.append(row)
                    max_columns = max(max_columns, len(row))
                except (csv.Error, StopIteration):
                    # If parsing fails, keep the original line
                    csv_rows.append([line])
                    max_columns = max(max_columns, 1)
            else:
                pass  # skip empty lines

        # Normalize all rows to have the same number of columns
        normalized_rows = []
        for row in csv_rows:
            normalized_row = row.copy()
            if len(normalized_row) < max_columns:
                # Pad with empty strings
                normalized_row.extend([""] * (max_columns - len(normalized_row)))
            elif len(normalized_row) > max_columns:
                # Truncate if somehow longer (shouldn't happen with max calculation)
                normalized_row = normalized_row[:max_columns]
            normalized_rows.append(normalized_row)

        # Convert back to CSV string
        output = StringIO()
        writer = csv.writer(output)
        writer.writerows(normalized_rows)
        output.seek(0)

        # Create a new NamedFileContents with the fixed content
        return NamedFileContents(
            contents=output,
            original_file_path=named_file_contents.original_file_path,
            encoding="utf-8",
        )



from __future__ import annotations

import os
import re
import traceback
import warnings
from xml.etree import ElementTree

# xml fromstring is vulnerable so defusedxml version is used instead
from defusedxml.ElementTree import fromstring

from allotropy.parsers.utils.values import assert_not_none, try_float_or_none


class StrictXmlElement:
    @classmethod
    def create_from_bytes(
        cls, data: bytes, mark_read: set[str] | None = None
    ) -> StrictXmlElement:
        return StrictXmlElement(fromstring(data), mark_read=mark_read)

    def __init__(
        self,
        element: ElementTree.Element,
        namespaces: dict[str, str] | None = None,
        mark_read: set[str] | None = None,
    ):
        self.element = element
        self.namespaces = namespaces or {}
        self.read_keys: set[str] = set()
        self.errored = False
        self.creation_stack = traceback.extract_stack()
        self.mark_read(mark_read or set())

    def __del__(self) -> None:
        if self.errored:
            return
        # NOTE: this will be turned on by default when all callers have been updated to pass the warning.
        # Only consider attributes as available keys, not child elements
        attribute_keys: set[str] = set()

        # Add attributes with "attr:" prefix
        for attr_name in self.element.attrib.keys():
            attribute_keys.add(f"attr:{attr_name}")

            # Also check if this is a namespaced attribute and add the ns_attr version
            for namespace_key, namespace_uri in self.namespaces.items():
                if attr_name.startswith(f"{{{namespace_uri}}}"):
                    field_name = attr_name.replace(f"{{{namespace_uri}}}", "")
                    attribute_keys.add(f"ns_attr:{namespace_key}:{field_name}")

        # Filter out attributes that have been read in either form (attr: or ns_attr:)
        unread_keys: set[str] = set()
        for key in attribute_keys:
            if key.startswith("attr:"):
                attr_name = key.removeprefix("attr:")

                # Check if this is a namespaced attribute that has been read via ns_attr
                is_namespaced_and_read = False
                for namespace_key, namespace_uri in self.namespaces.items():
                    if attr_name.startswith(f"{{{namespace_uri}}}"):
                        field_name = attr_name.replace(f"{{{namespace_uri}}}", "")
                        ns_attr_key = f"ns_attr:{namespace_key}:{field_name}"
                        if ns_attr_key in self.read_keys:
                            is_namespaced_and_read = True
                            break

                # Only consider unread if it hasn't been read via namespaced access and not directly read
                if key not in self.read_keys and not is_namespaced_and_read:
                    unread_keys.add(key)

            elif key.startswith("ns_attr:"):
                # Parse "ns_attr:namespace:field"
                parts = key.split(":", 2)
                if len(parts) == 3:
                    namespace_key, field = parts[1], parts[2]
                    if namespace_key in self.namespaces:
                        namespace_uri = self.namespaces[namespace_key]
                        full_attr_key = f"attr:{{{namespace_uri}}}{field}"

                        # Only consider unread if neither the ns_attr nor the full URI version has been read
                        if (
                            key not in self.read_keys
                            and full_attr_key not in self.read_keys
                        ):
                            unread_keys.add(key)

        # Only warn if we've read at least one key but haven't read all keys
        # This avoids warnings for elements that were never accessed at all
        if unread_keys and self.read_keys:
            if os.getenv("WARN_UNUSED_KEYS"):
                # Find the creation point in the stack (skip the StrictXmlElement.__init__ frame)
                creation_point = None
                for frame in reversed(self.creation_stack):
                    if (
                        frame.name != "__init__"
                        or "strict_xml_element.py" not in frame.filename
                    ):
                        creation_point = (
                            f"{frame.filename}:{frame.lineno} in {frame.name}"
                        )
                        break

                creation_info = (
                    f" (created at {creation_point})" if creation_point else ""
                )

                warnings.warn(
                    f"StrictXmlElement '{self.element.tag}' went out of scope without reading all keys{creation_info}, unread: {sorted(unread_keys)}.",
                    stacklevel=2,
                )

    def _get_all_available_keys(self) -> set[str]:
        """Get all available keys (attributes and child elements) from the XML element."""
        keys = set()

        # Add attributes with "attr:" prefix
        for attr_name in self.element.attrib.keys():
            keys.add(f"attr:{attr_name}")

            # Also check if this is a namespaced attribute and add the ns_attr version
            for namespace_key, namespace_uri in self.namespaces.items():
                if attr_name.startswith(f"{{{namespace_uri}}}"):
                    field_name = attr_name.replace(f"{{{namespace_uri}}}", "")
                    keys.add(f"ns_attr:{namespace_key}:{field_name}")

        # Add child elements with "element:" prefix
        child_names = set()
        for child in self.element:
            if child.tag not in child_names:
                keys.add(f"element:{child.tag}")
                child_names.add(child.tag)

        # Add text content if present
        if self.element.text and self.element.text.strip():
            keys.add("text:")

        return keys

    def _key_matches(self, match_key: str, key: str) -> bool:
        if key == match_key:
            return True
        # "++" can cause re.compile to fail. Since it is never a valid regex expression
        # it is safe to escape it to prevent the error.
        match_key = match_key.replace("++", r"\+\+")
        return bool(re.fullmatch(match_key, key))

    def _get_matching_keys(self, key_or_keys: str | set[str]) -> set[str]:
        """Get keys that match the given pattern(s)."""
        all_keys = self._get_all_available_keys()

        # Normalize keys by adding appropriate prefixes if they don't have them
        normalized_keys = set()
        for regex_key in key_or_keys if isinstance(key_or_keys, set) else {key_or_keys}:
            # If the key doesn't have a prefix and it's a simple attribute name,
            # automatically add the "attr:" prefix
            if not any(
                regex_key.startswith(prefix)
                for prefix in ["attr:", "ns_attr:", "element:", "text:"]
            ):
                # Check if this matches any actual attribute name (including namespaced ones)
                attr_key = f"attr:{regex_key}"
                matches_found = False

                # Check for exact match
                if any(k for k in all_keys if self._key_matches(attr_key, k)):
                    normalized_keys.add(attr_key)
                    matches_found = True

                # Check for namespaced matches (e.g., "schemaLocation" should match "attr:{namespace}schemaLocation")
                for key in all_keys:
                    if key.startswith("attr:"):
                        attr_name = key[5:]  # Remove "attr:" prefix
                        # Check if the attribute name ends with our regex_key (for namespaced attributes)
                        if (
                            attr_name.endswith(regex_key)
                            and "{" in attr_name
                            and "}" in attr_name
                        ):
                            # This is a namespaced attribute that ends with our key
                            normalized_keys.add(key)
                            matches_found = True

                if not matches_found:
                    # Keep the original key for other types of matches
                    normalized_keys.add(regex_key)
            else:
                normalized_keys.add(regex_key)

        return {
            matched
            for regex_key in normalized_keys
            for matched in [k for k in all_keys if self._key_matches(str(regex_key), k)]
        }

    def mark_read(self, key_or_keys: str | set[str]) -> None:
        """Mark specified keys as read."""
        self.read_keys |= self._get_matching_keys(key_or_keys)

    def mark_all_as_read(self) -> None:
        """Mark all available keys as read to avoid warnings."""
        self.read_keys = self._get_all_available_keys()

    def get_unread(
        self, regex: str | None = None, skip: set[str] | None = None
    ) -> dict[str, str | None]:
        """
        Get unread attributes from this element.

        Args:
            regex: Optional regex pattern to filter keys
            skip: Set of keys to skip

        Returns a dictionary with clean attribute names (prefixes removed):
        - attribute names without "attr:" prefix
        - namespaced attribute names without "ns_attr:" prefix
        """
        all_keys = self._get_all_available_keys()
        attribute_keys = self._get_all_attribute_keys()
        non_attribute_keys = all_keys - attribute_keys
        self._handle_skip_keys(skip)

        matching_keys_attribute = self._apply_regex_filter(attribute_keys, regex)
        matching_keys_non_attribute = self._apply_regex_filter(
            non_attribute_keys, regex
        )
        matching_keys = matching_keys_attribute | matching_keys_non_attribute

        unread_keys: dict[str, str | None] = {}
        processed_attrs: set[str] = set()

        for key in matching_keys - self.read_keys:
            if key.startswith("attr:"):
                self._process_attr_key(key, matching_keys, unread_keys, processed_attrs)
            elif key.startswith("ns_attr:"):
                self._process_ns_attr_key(key, unread_keys, processed_attrs)
            elif key.startswith("element:"):
                if self.should_process_element_key(key):
                    self._process_element_key(key, unread_keys)
                else:
                    continue

        self._mark_processed_keys_as_read(matching_keys, unread_keys)

        return dict(sorted(unread_keys.items()))

    def _get_all_attribute_keys(self) -> set[str]:
        """Get all attribute keys with appropriate prefixes."""
        attribute_keys = set()

        for attr_name in self.element.attrib.keys():
            attribute_keys.add(f"attr:{attr_name}")

            # Add namespaced versions
            for namespace_key, namespace_uri in self.namespaces.items():
                if attr_name.startswith(f"{{{namespace_uri}}}"):
                    field_name = attr_name.replace(f"{{{namespace_uri}}}", "")
                    attribute_keys.add(f"ns_attr:{namespace_key}:{field_name}")

        return attribute_keys

    def _handle_skip_keys(self, skip: set[str] | None) -> None:
        """Process and mark skip keys as read."""
        if skip:
            skip_keys = self._get_matching_keys(skip)
            self.read_keys |= skip_keys

    def _apply_regex_filter(
        self, attribute_keys: set[str], regex: str | None
    ) -> set[str]:
        """Apply regex filter to attribute keys if provided."""
        if regex:
            return {k for k in attribute_keys if self._key_matches(regex, k)}
        return attribute_keys

    def _process_attr_key(
        self,
        key: str,
        matching_keys: set[str],
        unread_keys: dict[str, str | None],
        processed_attrs: set[str],
    ) -> None:
        """Process an 'attr:' prefixed key."""
        attr_name = key[5:]  # Remove "attr:" prefix

        if self._is_attr_already_processed(attr_name, processed_attrs):
            return

        # Check if this should be skipped in favor of namespace form
        if self._should_skip_for_namespace_form(attr_name, matching_keys):
            return

        clean_key = self._get_clean_attribute_name(attr_name)
        unread_keys[clean_key] = self.element.get(attr_name)
        processed_attrs.add(attr_name)

    def _process_ns_attr_key(
        self, key: str, unread_keys: dict[str, str | None], processed_attrs: set[str]
    ) -> None:
        """Process a 'ns_attr:' prefixed key."""
        parts = key.split(":", 2)
        if len(parts) != 3:
            return

        namespace_key, field = parts[1], parts[2]
        if namespace_key not in self.namespaces:
            return

        namespace_uri = self.namespaces[namespace_key]
        full_attr_name = f"{{{namespace_uri}}}{field}"
        full_attr_key = f"attr:{full_attr_name}"

        # Check if already processed
        if (
            key in self.read_keys
            or full_attr_key in self.read_keys
            or full_attr_name in processed_attrs
        ):
            return

        clean_key = f"{namespace_key}:{field}"
        unread_keys[clean_key] = self.element.get(full_attr_name)
        processed_attrs.add(full_attr_name)

    def _process_element_key(
        self, key: str, unread_keys: dict[str, str | None]
    ) -> None:
        """Process an 'element:' prefixed key."""
        element_name = key.replace("element:", "")  # Remove "element:" prefix
        element = getattr(self.find_or_none(element_name), "element", None)
        if element is None:
            return
        value_for_key = element.text
        if value_for_key is not None:
            unread_keys[element_name] = value_for_key

    def should_process_element_key(self, key: str) -> bool:
        """
        Check if an XML element key should be processed.

        Returns False if:
        - The element has children
        - The element's text content appears to be XML formatted

        Returns True otherwise.
        """
        # Extract element name from key
        element_name = key.replace("element:", "")

        # Find the element
        element_wrapper = self.find_or_none(element_name)
        element = getattr(element_wrapper, "element", None)
        if element is None:
            return False

        # Check if element has children
        if len(element) > 0:
            return False

        # Check if element text looks like XML
        text = element.text
        if text:
            text = text.strip()
            if text:
                # Check for basic XML patterns
                if text.startswith("<") and text.endswith(">"):
                    return False

                # Check for XML-like patterns with regex
                xml_pattern = re.compile(r"<[^>]+>.*</[^>]+>", re.DOTALL)
                if xml_pattern.search(text):
                    return False

        return True

    def _is_attr_already_processed(
        self, attr_name: str, processed_attrs: set[str]
    ) -> bool:
        """Check if attribute has already been processed or read via namespaced access."""
        if attr_name in processed_attrs:
            return True

        # Check if read via ns_attr
        for namespace_key, namespace_uri in self.namespaces.items():
            if attr_name.startswith(f"{{{namespace_uri}}}"):
                field_name = attr_name.replace(f"{{{namespace_uri}}}", "")
                ns_attr_key = f"ns_attr:{namespace_key}:{field_name}"
                if ns_attr_key in self.read_keys:
                    return True
        return False

    def _should_skip_for_namespace_form(
        self, attr_name: str, matching_keys: set[str]
    ) -> bool:
        """Check if this attr should be skipped because namespace form is available."""
        for namespace_key, namespace_uri in self.namespaces.items():
            if attr_name.startswith(f"{{{namespace_uri}}}"):
                field_name = attr_name.replace(f"{{{namespace_uri}}}", "")
                ns_attr_form = f"ns_attr:{namespace_key}:{field_name}"
                if ns_attr_form in matching_keys:
                    return True
        return False

    def _get_clean_attribute_name(self, attr_name: str) -> str:
        """Convert attribute name to clean format, handling namespaces."""
        for ns_key, ns_uri in self.namespaces.items():
            if attr_name.startswith(f"{{{ns_uri}}}"):
                field_name = attr_name.replace(f"{{{ns_uri}}}", "")
                # Special handling for XMLSchema-instance namespace
                if ns_key == "xsi":
                    return field_name
                else:
                    return f"{ns_key}:{field_name}"
        return attr_name

    def _mark_processed_keys_as_read(
        self, matching_keys: set[str], unread_keys: dict[str, str | None]
    ) -> None:
        """Mark the processed keys as read."""
        if unread_keys:
            keys_to_mark = {
                key
                for key in matching_keys - self.read_keys
                if key.startswith(("attr:", "ns_attr:"))
            }
            self.read_keys |= keys_to_mark

    def find_or_none(self, name: str) -> StrictXmlElement | None:
        self.read_keys.add(f"element:{name}")
        element = self.element.find(name, self.namespaces)
        return (
            StrictXmlElement(element, self.namespaces) if element is not None else None
        )

    def find(self, name: str) -> StrictXmlElement:
        try:
            return assert_not_none(
                self.find_or_none(name),
                msg=f"Unable to find '{name}' in xml file contents",
            )
        except Exception:
            self.errored = True
            raise

    def recursive_find_or_none(self, names: list[str]) -> StrictXmlElement | None:
        if len(names) == 0:
            return self
        name, *sub_names = names
        if element := self.find_or_none(name):
            # Mark all attributes as read for intermediate elements
            # (not for the final element since that's the one we want to return)
            if sub_names:
                element.get_unread()
            return element.recursive_find_or_none(sub_names)
        return None

    def recursive_find(self, names: list[str]) -> StrictXmlElement:
        if len(names) == 0:
            return self
        name, *sub_names = names
        element = self.find(name)
        if sub_names:
            element.get_unread()
        return element.recursive_find(sub_names)

    def findall(self, name: str) -> list[StrictXmlElement]:
        self.read_keys.add(f"element:{name}")
        return [
            StrictXmlElement(element, self.namespaces)
            for element in self.element.findall(name, self.namespaces)
        ]

    def get_attr_or_none(self, name: str) -> str | None:
        self.read_keys.add(f"attr:{name}")
        value = self.element.get(name)
        return None if value is None else str(value)

    def get_attr(self, name: str) -> str:
        try:
            return assert_not_none(
                self.get_attr_or_none(name),
                msg=f"Unable to find '{name}' in xml file contents",
            )
        except Exception:
            self.errored = True
            raise

    def parse_text_or_none(self) -> StrictXmlElement | None:
        if (text := self.get_text_or_none()) is None:
            return None
        try:
            return StrictXmlElement(fromstring(text), self.namespaces)
        except ElementTree.ParseError:
            return None

    def parse_text(self, name: str) -> StrictXmlElement:
        try:
            return assert_not_none(
                self.parse_text_or_none(),
                msg=f"Unable to parse text from xml tag '{name}' as valid xml content",
            )
        except Exception:
            self.errored = True
            raise

    def get_text_or_none(self) -> str | None:
        if self.element.text is not None:
            self.read_keys.add("text:")
        return self.element.text

    def get_text(self, name: str) -> str:
        try:
            return assert_not_none(
                self.get_text_or_none(),
                msg=f"Unable to find valid string from xml tag '{name}'",
            )
        except Exception:
            self.errored = True
            raise

    def get_float_or_none(self) -> float | None:
        return try_float_or_none(self.get_text_or_none())

    def get_float(self, name: str) -> float:
        try:
            return assert_not_none(self.get_float_or_none(), name)
        except Exception:
            self.errored = True
            raise

    def get_sub_float_or_none(self, name: str) -> float | None:
        if element := self.find_or_none(name):
            return element.get_float_or_none()
        return None

    def get_sub_text_or_none(self, name: str) -> str | None:
        if element := self.find_or_none(name):
            return element.get_text_or_none()
        return None

    # namespace-specific methods
    def get_namespaced_attr_or_none(self, namespace_key: str, field: str) -> str | None:
        self.read_keys.add(f"ns_attr:{namespace_key}:{field}")
        if namespace_key not in self.namespaces:
            return None
        return self.element.get(f"{{{self.namespaces.get(namespace_key)}}}{field}")

    def get_namespaced_attr(self, namespace_key: str, field: str) -> str:
        try:
            return assert_not_none(
                self.get_namespaced_attr_or_none(namespace_key, field),
                msg=f"Unable to find '{namespace_key}:{field}' in xml file contents",
            )
        except Exception:
            self.errored = True
            raise



DISPLAY_NAME = "Cytiva Biacore Insight"
DEVICE_IDENTIFIER = "Biacore"
DEVICE_TYPE = "Binding Affinity Analyzer"
DETECTION_TYPE = "Surface Plasmon Resonance"
PRODUCT_MANUFACTURER = "Cytiva"

COMPARTMENT_TEMP_REGEX = r"Sample compartment temperature (\d+),.*"



from enum import Enum
import math
from pathlib import Path

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueSecondTime,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TStatisticDatumRole,
)
from allotropy.allotrope.schema_mappers.adm.flow_cytometry.benchling._2025._03.flow_cytometry import (
    CompensationMatrix,
    CompensationMatrixGroup,
    DataRegion,
    Measurement,
    MeasurementGroup,
    Metadata,
    Population,
    Statistic,
    StatisticDimension,
    Vertex,
)
from allotropy.exceptions import AllotropeParsingError
from allotropy.parsers.bd_biosciences_facsdiva import constants
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.strict_xml_element import StrictXmlElement
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import try_float_or_none, try_int_or_none

# Map of FACSDiva field names to statistic datum roles and units
FACSDIVA_STATISTIC_MAP = {
    # Fluorescence statistics
    "Median": {"role": TStatisticDatumRole.median_role.value, "unit": "RFU"},
    "CV": {
        "role": TStatisticDatumRole.coefficient_of_variation_role.value,
        "unit": "%",
    },
    "Mean": {"role": TStatisticDatumRole.arithmetic_mean_role.value, "unit": "RFU"},
    "Geometric Mean": {
        "role": TStatisticDatumRole.geometric_mean_role.value,
        "unit": "RFU",
    },
    "SD": {
        "role": TStatisticDatumRole.standard_deviation_role.value,
        "unit": "(unitless)",
    },
    "sum of squares": {
        "role": TStatisticDatumRole.sum_of_squares_role.value,
        "unit": "RFU",
    },
}

FLUORESCENCE_FIELDS = [
    "Median",
    "CV",
    "Mean",
    "Geometric Mean",
    "SD",
    "sum of squares",
]

COUNT_FIELDS = [
    "fj.stat.freqofparent",
    "fj.stat.freqofgrandparent",
    "fj.stat.freqoftotal",
]

# These channels are not part of the compensation matrix
EXCLUDED_CHANNELS = [
    "FSC-A",
    "FSC-H",
    "FSC-W",
    "SSC-A",
    "SSC-H",
    "SSC-W",
    "Time",
]

FIELDS_TO_SKIP = {
    "$OP",  # Already captured as analyst
}

# Fields for device control document
DEVICE_CONTROL_FIELDS = {
    "CST SETUP STATUS",
    "CST BEADS LOT ID",
    "CYTOMETER CONFIG NAME",
    "CYTOMETER CONFIG CREATE DATE",
    "CST SETUP DATE",
    "CST BASELINE DATE",
    "CST BEADS EXPIRED",
}

MEASUREMENT_AGGREGATE_FIELDS = {
    "$BTIM",
}


def _filter_unread_data(unread_data: dict[str, str | None]) -> dict[str, str | None]:
    """Filter out empty, null, or whitespace-only values from unread data."""
    return {
        key: value
        for key, value in unread_data.items()
        if not (
            (value == "")
            or (value is None)
            or (isinstance(value, str) and value.strip() == "")
        )
    }


def _separate_custom_fields(
    custom_info: dict[str, str]
) -> tuple[dict[str, str], dict[str, str], dict[str, str]]:
    device_control_info = {}
    measurement_aggregate_info = {}
    other_info = {}

    for key, value in custom_info.items():
        if key in FIELDS_TO_SKIP:
            continue
        elif key in DEVICE_CONTROL_FIELDS:
            device_control_info[key] = value
        elif key in MEASUREMENT_AGGREGATE_FIELDS:
            measurement_aggregate_info[key] = value
        else:
            other_info[key] = value

    return device_control_info, measurement_aggregate_info, other_info


class RegionType(Enum):
    POLYGON = "POLYGON"
    RECTANGLE = "RECTANGLE"
    BINNER = "BINNER"


def create_metadata(root_element: StrictXmlElement, file_path: str) -> Metadata:
    equipment_serial_number = None
    model_number = None
    all_custom_info = {}

    tube = root_element.recursive_find_or_none(["experiment", "specimen", "tube"])
    if tube:
        equipment_serial_number = tube.find_or_none("data_instrument_serial_number")
        model_number = tube.find_or_none("data_instrument_name")
        tube.mark_read("name")

    software_version = root_element.get_attr_or_none("version")

    # Collect unread data from root element
    root_unread = root_element.get_unread(
        skip={
            "version",
            "release_version",
        }
    )
    if root_unread:
        filtered_root_unread = _filter_unread_data(root_unread)
        all_custom_info.update(filtered_root_unread)

    return Metadata(
        file_name=Path(file_path).name,
        unc_path=file_path,
        device_identifier=constants.DEVICE_IDENTIFIER,
        model_number=model_number.get_text_or_none() if model_number else None,
        equipment_serial_number=equipment_serial_number.get_text_or_none()
        if equipment_serial_number
        else None,
        data_system_instance_identifier=NOT_APPLICABLE,
        software_name=constants.SOFTWARE_NAME,
        software_version=software_version,
        asm_file_identifier=Path(file_path).with_suffix(".json").name,
        custom_info=all_custom_info if all_custom_info else None,
    )


def _extract_statistics_from_calculations(
    tube: StrictXmlElement,
    gate_name: str,
) -> list[Statistic] | None:
    """
    Extract statistics from calculation schedules for a specific gate.

    Args:
        tube: The tube element containing statistics
        gate_name: The name of the gate to extract statistics for

    Returns:
        list[Statistic] | None: List of statistics or None if no statistics found
    """
    statistics_element = tube.find_or_none("statistics_calculations")
    if not statistics_element:
        return None

    fluorescence_dimensions = []
    count_dimensions = []

    for calc_schedule in statistics_element.findall("calculation_schedule"):
        gate = calc_schedule.get_attr_or_none("gate")
        if gate != gate_name:
            calc_schedule.get_unread()
            continue

        parameter = calc_schedule.get_attr_or_none("parameter")

        for calculation in calc_schedule.findall("calculation"):
            formula = calculation.get_attr_or_none("formula")
            value_str = calculation.get_attr_or_none("value")

            if not (formula and value_str):
                continue

            value = try_float_or_none(value_str)
            if value is None or math.isnan(value):
                continue
            if formula == "count":
                # Skip count statistics as it is being reported in the population
                continue

            stat_info = FACSDIVA_STATISTIC_MAP.get(formula)
            if stat_info is None:
                continue

            dimension = StatisticDimension(
                dimension_identifier=parameter,
                value=value,
                unit=stat_info["unit"],
                has_statistic_datum_role=stat_info["role"],
            )

            if formula in FLUORESCENCE_FIELDS:
                fluorescence_dimensions.append(dimension)
            elif formula in COUNT_FIELDS:
                count_dimensions.append(dimension)

    statistics = []

    if fluorescence_dimensions:
        statistics.append(
            Statistic(
                statistical_feature="fluorescence",
                statistic_dimension=fluorescence_dimensions,
            )
        )

    if count_dimensions:
        statistics.append(
            Statistic(statistical_feature="count", statistic_dimension=count_dimensions)
        )

    return statistics if statistics else None


def _create_data_regions(tube: StrictXmlElement) -> list[DataRegion]:
    """
    Create data regions from gates in the tube.

    Args:
        tube: The tube element containing the gates

    Returns:
        list[DataRegion]: List of DataRegion objects
    """
    data_regions = []

    gates_element = tube.find_or_none("gates")
    if not gates_element:
        return []

    fullname_to_region = {}
    region_to_data_region = {}

    # First pass: collect gate information
    for gate in gates_element.findall("gate"):
        # Skip All Events full name (no region)
        gate_type = gate.get_attr_or_none("type")
        fullname = gate.get_attr_or_none("fullname")

        if gate_type == "EventSource_Classifier":
            continue

        region = gate.find_or_none("region")
        if not region:
            continue

        if not fullname:
            continue

        region_name = region.get_attr_or_none("name")
        if region_name:
            fullname_to_region[fullname] = region_name

        region.mark_read(
            {
                "type",
                "xparm",
                "yparm",
            }
        )

    # Second pass: create data regions
    for gate in gates_element.findall("gate"):
        gate_type = gate.get_attr_or_none("type")
        fullname = gate.get_attr_or_none("fullname")

        if gate_type == "EventSource_Classifier":
            continue

        region = gate.find_or_none("region")
        if not region:
            continue

        if not fullname:
            continue

        region_name = region.get_attr_or_none("name")
        region_type_str = region.get_attr_or_none("type")
        x_param = region.get_attr_or_none("xparm")
        y_param = region.get_attr_or_none("yparm")

        region_type = None
        if region_type_str and "_REGION" in region_type_str:
            base_type = region_type_str.replace("_REGION", "")
            region_type = {e.name: e.value for e in RegionType}.get(base_type)

        # Find parent region identifier using fullname hierarchy
        parent_data_region_id = None
        if "\\" in fullname:
            parent_fullname = fullname.rsplit("\\", 1)[0]
            # Skip All Events as parent (no region)
            if parent_fullname != "All Events":
                parent_data_region_id = fullname_to_region.get(parent_fullname)

        # Extract vertices
        vertices = []
        points_element = region.find_or_none("points")
        if points_element:
            for point in points_element.findall("point"):
                x_coord = try_float_or_none(point.get_attr_or_none("x"))
                y_coord = try_float_or_none(point.get_attr_or_none("y"))

                if x_coord is not None and y_coord is not None:
                    x_unit = (
                        TQuantityValueSecondTime.unit
                        if x_param and "Time" in x_param
                        else TQuantityValueRelativeFluorescenceUnit.unit
                    )
                    y_unit = (
                        TQuantityValueSecondTime.unit
                        if y_param and "Time" in y_param
                        else TQuantityValueRelativeFluorescenceUnit.unit
                    )

                    vertices.append(
                        Vertex(
                            x_coordinate=x_coord,
                            y_coordinate=y_coord,
                            x_unit=x_unit,
                            y_unit=y_unit,
                        )
                    )

        data_region = DataRegion(
            region_data_identifier=region_name,
            region_data_type=region_type,
            parent_data_region_identifier=parent_data_region_id,
            x_coordinate_dimension_identifier=x_param,
            y_coordinate_dimension_identifier=y_param,
            vertices=vertices if vertices else None,
        )

        data_regions.append(data_region)
        if region_name:
            region_to_data_region[region_name] = data_region

    return data_regions


def _create_populations(tube: StrictXmlElement) -> list[Population]:
    """
    Create populations from gates in the tube.

    Args:
        tube: The tube element containing the gates

    Returns:
        list[Population]: List containing the root population with nested subpopulations
    """
    gates_element = tube.find_or_none("gates")
    if not gates_element:
        return []

    fullname_to_gate = {}
    gate_to_children: dict[str, list[str]] = {}
    for gate in gates_element.findall("gate"):
        fullname = gate.get_attr_or_none("fullname")
        if not fullname:
            continue

        fullname_to_gate[fullname] = gate

        parent_elem = gate.find_or_none("parent")
        if parent_elem and parent_elem.get_text_or_none():

            # Build parent path using fullname structure
            if "\\" in fullname:
                parent_path = fullname.rsplit("\\", 1)[0]
                if parent_path not in gate_to_children:
                    gate_to_children[parent_path] = []
                gate_to_children[parent_path].append(fullname)

    def _build_population_tree(full_name: str) -> Population | None:
        _gate = fullname_to_gate.get(full_name)
        if not _gate:
            return None

        pop_id = random_uuid_str()

        name_elem = _gate.find_or_none("name")
        name = name_elem.get_text_or_none() if name_elem else None

        count_elem = _gate.find_or_none("num_events")
        count = try_int_or_none(count_elem.get_text_or_none()) if count_elem else None

        region = _gate.find_or_none("region")
        region_id = region.get_attr_or_none("name") if region else None

        if region:
            region.mark_read(
                {
                    "type",
                    "xparm",
                    "yparm",
                }
            )
        _gate.mark_read(
            {
                "type",
            }
        )

        sub_populations = []
        if full_name in gate_to_children:
            for child_fullname in gate_to_children[full_name]:
                child_pop = _build_population_tree(child_fullname)
                if child_pop:
                    updated_child_pop = Population(
                        population_identifier=child_pop.population_identifier,
                        parent_population_identifier=pop_id,
                        written_name=child_pop.written_name,
                        data_region_identifier=child_pop.data_region_identifier,
                        count=child_pop.count,
                        sub_populations=child_pop.sub_populations,
                        statistics=child_pop.statistics,
                        custom_info=child_pop.custom_info,
                    )
                    sub_populations.append(updated_child_pop)

        # Create this population with its children
        return Population(
            population_identifier=pop_id,
            parent_population_identifier=None,
            written_name=name,
            data_region_identifier=region_id,
            count=count,
            sub_populations=sub_populations if sub_populations else None,
            statistics=_extract_statistics_from_calculations(tube, full_name),
            custom_info=None,
        )

    # Find the root population (All Events) and build the tree
    root_gates = []
    for gate in gates_element.findall("gate"):
        gate_type = gate.get_attr_or_none("type")
        if gate_type == "EventSource_Classifier":
            root_gates.append(gate)
        gate.mark_read(
            {
                "fullname",
            }
        )

    if not root_gates:
        return []

    root_fullname = root_gates[0].get_attr_or_none("fullname")
    if not root_fullname:
        return []

    root_population = _build_population_tree(root_fullname)
    return [root_population] if root_population else []


def _create_compensation_matrix_groups(
    tube: StrictXmlElement,
) -> list[CompensationMatrixGroup] | None:
    """
    Create compensation matrix groups from instrument settings.

    Args:
        tube: The tube element containing instrument settings

    Returns:
        list[CompensationMatrixGroup] | None: List of compensation matrix groups or None
    """
    instrument_settings = tube.find_or_none("instrument_settings")
    if not instrument_settings:
        return None
    # Get fluorescent parameters
    parameters = []
    for param in instrument_settings.findall("parameter"):
        name = param.get_attr_or_none("name")
        if name and name not in EXCLUDED_CHANNELS:
            parameters.append(param)
        else:
            param.get_unread()

    if not parameters:
        return None

    result = []
    for param in parameters:
        dimension_id = param.get_attr_or_none("name")
        compensation_element = param.find_or_none("compensation")

        if not (dimension_id and compensation_element):
            continue

        matrices = []
        # For each coefficient, create a compensation matrix entry
        for i, coef_elem in enumerate(
            compensation_element.findall("compensation_coefficient")
        ):
            coef_value = try_float_or_none(coef_elem.get_text_or_none())
            if coef_value is None:
                continue

            # Get the corresponding parameter for this coefficient
            target_param = parameters[i]
            target_dimension_id = target_param.get_attr_or_none("name")

            if not target_dimension_id:
                continue

            matrices.append(
                CompensationMatrix(
                    dimension_identifier=target_dimension_id,
                    compensation_value=coef_value,
                )
            )
        param.mark_read(
            "type",
        )

        result.append(
            CompensationMatrixGroup(
                dimension_identifier=dimension_id,
                compensation_matrices=matrices if matrices else None,
            )
        )
    instrument_settings.mark_read(
        {
            "name",
            "template",
        }
    )

    return result if result else None


def _process_tube(
    tube: StrictXmlElement, specimen_name: str | None, data_processing_time: str | None
) -> tuple[list[Measurement], dict[str, str]]:
    """
    Process a tube element to create measurements.

    Args:
        tube: The tube element to process

    Returns:
        tuple: (measurements, measurement_aggregate_custom_info)
    """
    tube_name = tube.get_attr("name")

    processed_data_id = random_uuid_str()

    # Extract custom info from keywords
    custom_info = {}
    keywords_element = tube.find_or_none("keywords")
    if keywords_element:
        for keyword in keywords_element.findall("keyword"):
            keyword_name = keyword.get_attr_or_none("name")
            keyword.mark_read("type")

            if keyword_name:
                value_element = keyword.find_or_none("value")
                if value_element:
                    keyword_value = value_element.get_text_or_none()
                    if keyword_value:
                        custom_info[keyword_name] = keyword_value

    (
        device_control_info,
        measurement_aggregate_info,
        other_info,
    ) = _separate_custom_fields(custom_info)

    # Create populations and data regions
    populations = _create_populations(tube)
    data_regions = _create_data_regions(tube)
    measurement = Measurement(
        measurement_identifier=random_uuid_str(),
        sample_identifier=tube_name,
        batch_identifier=specimen_name,
        device_type=constants.DEVICE_TYPE,
        data_processing_time=data_processing_time,
        processed_data_identifier=processed_data_id,
        populations=populations,
        data_regions=data_regions,
        device_control_custom_info=device_control_info if device_control_info else None,
        custom_info=other_info if other_info else None,
    )

    return [measurement], measurement_aggregate_info


def create_measurement_groups(root_element: StrictXmlElement) -> list[MeasurementGroup]:
    """
    Create measurement groups from FACSDiva XML.

    Args:
        root_element: The root element of the FACSDiva XML file

    Returns:
        list[MeasurementGroup]: List of measurement groups
    """

    result = []
    experiment = root_element.find_or_none("experiment")
    if not experiment:
        msg = "No experiment found in the XML file."
        raise AllotropeParsingError(msg)

    measurement_time = experiment.recursive_find_or_none(["specimen", "tube", "date"])
    experiment_identifier = experiment.get_attr_or_none("name")
    analyst = experiment.recursive_find_or_none(["owner_name"])
    export_time = experiment.find_or_none("export_time")
    # Process specimens and tubes
    specimens = experiment.findall("specimen")
    for specimen in specimens:
        specimen_name = specimen.get_attr_or_none("name")
        tubes = specimen.findall("tube")
        for tube in tubes:
            experimental_data_identifier = tube.find_or_none("data_filename")
            measurements, measurement_aggregate_custom_info = _process_tube(
                tube,
                specimen_name,
                export_time.get_text_or_none() if export_time else None,
            )

            compensation_matrix_groups = _create_compensation_matrix_groups(tube)
            measurement_group = MeasurementGroup(
                experimental_data_identifier=experimental_data_identifier.get_text_or_none()
                if experimental_data_identifier
                else None,
                measurement_time=measurement_time.get_text_or_none()
                if measurement_time
                else None,
                analyst=analyst.get_text_or_none() if analyst else None,
                compensation_matrix_groups=compensation_matrix_groups,
                measurements=measurements,
                experiment_identifier=experiment_identifier,
                measurement_aggregate_custom_info=measurement_aggregate_custom_info
                if measurement_aggregate_custom_info
                else None,
            )

            result.append(measurement_group)

    return result
